import json
import random
import shutil
import string
import zipfile
from dataclasses import dataclass
from enum import Enum
from itertools import combinations
from pathlib import Path
from typing import Any, List, Optional, Tuple, TypedDict, Union

import numpy as np
import numpy.typing as npt
import pandas as pd

from .clustering.common import ClusteringContext, Clusters, Entropy1Dim
from .clustering.features import FeatureSelectionResult, select_features_ml
from .clustering.measures import DependenceMeasures, measure_all
from .clustering.solver import _do_solve, solve_with_features
from .clustering.strategy import NoClustering
from .common import (
    AnonymizationParams,
    BucketizationParams,
    ColumnId,
    FlatteningInterval,
    SuppressionParams,
)
from .stitcher import stitch
from .synthesizer import Synthesizer


def _shrink_matrix(matrix: npt.NDArray, comb: tuple[int, ...]) -> npt.NDArray:
    return matrix[np.ix_(comb, comb)]


def _shrink_entropy_1dim(entropy_1dim: npt.NDArray[np.float64], comb: tuple[int, ...]) -> Entropy1Dim:
    return entropy_1dim[list(comb)]


def _parquet_reader(file_path: Path) -> pd.DataFrame:
    try:
        return pd.read_parquet(file_path)
    except Exception as e:
        raise IOError(f"Failed to read Parquet file {file_path}: {e}") from e


class StitchDict(TypedDict):
    left: List[str]
    right: List[str]
    owner: str


class StitchRecord:
    def __init__(self) -> None:
        self._stitch_record: List[StitchDict] = []

    def add(self, left: list[str], right: list[str], owner: str) -> None:
        stitch_entry: StitchDict = {"left": left, "right": right, "owner": owner}
        self._stitch_record.append(stitch_entry)

    def read(self) -> List[StitchDict]:
        return self._stitch_record

    def reset(self) -> None:
        self._stitch_record = []


class BlobZipper:
    def __init__(self, blob_name: str, path_to_dir: Path, path_to_blob_dir: Path) -> None:
        self.suffix = ".sdxblob.zip"
        self.blob_name = blob_name
        self.path_to_dir = path_to_dir
        self.path_to_blob_dir = path_to_blob_dir

    def zip_blob(self) -> None:
        zip_file_path = self.path_to_dir.joinpath(f"{self.blob_name}.sdxblob.zip")
        shutil.make_archive(str(zip_file_path.with_suffix("")), "zip", str(self.path_to_blob_dir))

    def unzip_blob(self) -> None:
        zip_file_path = self.path_to_dir.joinpath(f"{self.blob_name}.sdxblob.zip")
        if zip_file_path.exists():
            try:
                with zipfile.ZipFile(zip_file_path, "r") as zip_ref:
                    zip_ref.extractall(self.path_to_blob_dir)
            except zipfile.BadZipFile:
                print(f"Error: The file {zip_file_path} is not a valid ZIP file.")
            except zipfile.LargeZipFile:
                print(f"Error: The file {zip_file_path} requires ZIP64 functionality but it is not enabled.")
            except Exception as e:
                print(f"An unexpected error occurred while extracting {zip_file_path}: {e}")
        else:
            raise FileNotFoundError(f"Zip file {zip_file_path} does not exist.")


class ClusterParams:
    def __init__(
        self,
        sample_size: int = 1000,
        default_max_weight: float = 15.0,
        ml_max_weight: float = 15.0,
        merge_threshold: float = 0.1,
        solver_alpha: float = 1e-2,
        max_cluster_size: int = 10,  # This is effectively infinite
    ):
        self.sample_size = sample_size
        self.default_max_weight = default_max_weight
        self.ml_max_weight = ml_max_weight
        self.merge_threshold = merge_threshold
        self.solver_alpha = solver_alpha
        self.max_cluster_size = max_cluster_size


@dataclass(frozen=True)
class FeaturesContext:
    main_column: ColumnId
    main_features: list[ColumnId]
    entropy_1dim: Entropy1Dim


class BlobFiles(Enum):
    VERSION = "_meta_version.json"
    DEPENDENCE_MATRIX = "_meta_dependence_matrix.npy"
    ENTROPY_1DIM = "_meta_entropy_1dim.npy"
    ANONYMIZATION_PARAMS = "_meta_anonymization_params.json"
    BUCKETIZATION_PARAMS = "_meta_bucketization_params.json"
    CLUSTER_PARAMS = "_meta_cluster_params.json"
    COLUMNS = "_meta_columns.json"
    FEATURES = "_meta_features.json"


CatalogValueType = dict[str, Any]

CatalogType = dict[Tuple[str, ...], CatalogValueType]


class Catalog(object):
    def __init__(self) -> None:
        self.catalog: CatalogType = {}

    def is_valid_key(self, columns_tuple: Tuple[str, ...]) -> bool:
        return columns_tuple in self.catalog

    def add(self, columns_tuple: Tuple[str, ...], df: Optional[pd.DataFrame], file_path: Path) -> None:
        self.catalog[columns_tuple] = {"df": df, "file_path": file_path}

    def keys(self) -> List[Tuple[str, ...]]:
        return list(self.catalog.keys())

    def read(self, columns_tuple: Tuple[str, ...]) -> pd.DataFrame:
        if self.catalog[columns_tuple]["df"] is not None:
            return self.catalog[columns_tuple]["df"]
        return _parquet_reader(self.catalog[columns_tuple]["file_path"])


class SyndiffixBlob(object):
    def __init__(
        self,
        blob_name: str,
        path_to_dir: Optional[Union[str, Path]] = None,
        force: Optional[bool] = False,
    ) -> None:
        self.unsafe_rng = random.Random(0)
        self.blob_name = blob_name
        self.force = force
        # This is where the zipped blob is
        self.path_to_dir: Path
        self.blob_dir_name: str
        self.col_names_all: list[str]
        self.cluster_params = ClusterParams()

        # This is where the unzipped blob files go
        self.path_to_blob_dir: Path

        self.features: dict[str, FeatureSelectionResult]
        self.measures: DependenceMeasures
        self.context: ClusteringContext
        self.anonymization_params: AnonymizationParams
        self.bucketization_params: BucketizationParams

        self._make_paths(path_to_dir)
        self.bzip = BlobZipper(blob_name, self.path_to_dir, self.path_to_blob_dir)

    def _features_context(self, comb: tuple[int, ...], columns: tuple[str, ...], target_column: str) -> FeaturesContext:
        features_all = self.features[target_column]
        main_features = []
        # This algorithm is a bit funky, because features_all was computed from the complete table,
        # while here we are trying to interpret it relative to a partial table. This needs more thought,
        for feature in features_all.k_features:
            if feature in columns:
                feature_id = ColumnId(columns.index(feature))
                main_features.append(feature_id)
        return FeaturesContext(
            main_column=ColumnId(columns.index(target_column)),
            main_features=main_features,
            entropy_1dim=_shrink_entropy_1dim(self.measures.entropy_1dim, comb),
        )

    def _clustering_context(self, comb: tuple[int, ...]) -> ClusteringContext:
        # the index values in comb are relative to the all columns (self.col_names_all)
        # However, the indices in the returned clustering_context are relative to
        # the data structures in clustering_context
        dependency_matrix = _shrink_matrix(self.measures.dependency_matrix, comb)
        entropy_1dim = _shrink_entropy_1dim(self.measures.entropy_1dim, comb)
        total_per_column = [sum(dependency_matrix[i, j] for j in range(len(comb)) if i != j) for i in range(len(comb))]
        return ClusteringContext(
            dependency_matrix=dependency_matrix,
            entropy_1dim=entropy_1dim,
            total_dependence_per_column=total_per_column,
            total_dependence=sum(total_per_column),
            anonymization_params=self.anonymization_params,
            bucketization_params=self.bucketization_params,
            rng=self.unsafe_rng,
            main_column=None,
        )

    def _make_paths(self, path_to_dir: Optional[Union[str, Path]]) -> None:
        if path_to_dir is None:
            self.path_to_dir = Path.cwd()
        else:
            self.path_to_dir = Path(path_to_dir)
        if not self.path_to_dir.is_dir():
            raise NotADirectoryError(f"The path {self.path_to_dir} is not an existing directory.")
        self.blob_dir_name = f".sdx_blob_{self.blob_name}"
        self.path_to_blob_dir = self.path_to_dir.joinpath(self.blob_dir_name)
        self.path_to_blob_dir.mkdir(parents=True, exist_ok=True)

    def _data_filename(self, columns: list[str]) -> str:
        columns.sort()
        name = self.blob_name + f".col{len(columns)}."
        if len(columns) <= 10:
            chars_per_col = int(30 / len(columns)) + 1
            for col in columns:
                # substitute whitespace with underscore
                col = col.replace(" ", "_")
                illegal_chars = [
                    "<",
                    ">",
                    ":",
                    '"',
                    "/",
                    "\\",
                    "|",
                    "?",
                    "*",
                ]
                for c in illegal_chars:
                    col = col.replace(c, "_")
                name += col[:chars_per_col] + "_"
            # strip off the last underscore
            name = name[:-1]
        seed = self.blob_name + "__".join(columns)
        random.seed(seed)
        # append random alphanumeric characters to ensure that file name is unique
        name += "." + "".join(random.choices(string.ascii_lowercase + string.digits, k=4))
        return name

    def _get_column_names_from_indices(self, comb: Union[list[ColumnId], tuple[ColumnId, ...]]) -> tuple[str, ...]:
        return tuple([self.col_names_all[i] for i in comb])


class SyndiffixBlobBuilder(SyndiffixBlob):
    def __init__(
        self,
        blob_name: str,
        path_to_dir: Optional[Union[str, Path]] = None,
        force: Optional[bool] = False,
        sample_size: Optional[int] = None,
        default_max_weight: Optional[float] = None,
        ml_max_weight: Optional[float] = None,
        merge_threshold: Optional[float] = None,
        solver_alpha: Optional[float] = None,
        max_cluster_size: Optional[int] = None,
    ) -> None:
        super().__init__(blob_name, path_to_dir, force=force)  # Call the base class initializer
        self.write_version: int = 1

        if sample_size is not None:
            self.cluster_params.sample_size = sample_size
        if default_max_weight is not None:
            self.cluster_params.default_max_weight = default_max_weight
        if ml_max_weight is not None:
            self.cluster_params.ml_max_weight = ml_max_weight
        if merge_threshold is not None:
            self.cluster_params.merge_threshold = merge_threshold
        if solver_alpha is not None:
            self.cluster_params.solver_alpha = solver_alpha
        if max_cluster_size is not None:
            self.cluster_params.max_cluster_size = max_cluster_size

        self.df_raw: pd.DataFrame
        self.pids: Optional[pd.DataFrame]
        self.measures: DependenceMeasures
        self.context: ClusteringContext
        self.anonymization_params: AnonymizationParams
        self.bucketization_params: BucketizationParams

    def write(
        self,
        df_raw: pd.DataFrame,
        pids: Optional[pd.DataFrame] = None,
    ) -> None:
        """
        Parameters:
            df_raw: The raw data to be synthesized in a blob
            blob_name: The root name of the generated blob (will add version and .zip)
            pids: The PIDs to be used in the synthesis
            path_to_dir: The place to put the blob. Defaults to current directory
            sample_size, max_weight, merge_threshold, and solver_alpha control clustering
        """
        self.df_raw = df_raw
        self.pids = pids

        self._write_version()
        # Build a synthesizer that synthesizes each column independently of the others
        syn = Synthesizer(self.df_raw, pids=self.pids, clustering=NoClustering())
        self.col_names_all = list(syn.forest.columns)
        self.anonymization_params = syn.forest.anonymization_params
        self._write_anonymization_params()
        self.bucketization_params = syn.forest.bucketization_params
        self._write_bucketization_params()
        self._write_cluster_params()
        # Build the 1-dim tables
        for comb in combinations(range(len(self.col_names_all)), 1):
            # By forcing the initial cluster to be the columns we want, we trick syn.sample()
            syn.clusters = Clusters(
                initial_cluster=[ColumnId(comb[0])],
                derived_clusters=[],
            )
            df_1col = syn.sample()
            self._parquet_writer(df_1col, f"{self._data_filename(list(df_1col.columns))}.parquet")

        self._compute_ml_features()
        self._write_features()
        self._write_column_names()
        # Build all the 2-dim tables
        for comb in combinations(range(len(self.col_names_all)), 2):
            syn.clusters = Clusters(
                initial_cluster=[ColumnId(comb[0]), ColumnId(comb[1])],
                derived_clusters=[],
            )
            df_2col = syn.sample()
            self._parquet_writer(df_2col, f"{self._data_filename(list(df_2col.columns))}.parquet")
        # measure_all gives us the pairwise dependence measures and the 1dim entropy measures
        self.measures = measure_all(syn.forest)
        self._write_measures()
        if self.cluster_params.max_cluster_size > 2:
            # Build 3-dim and larger tables, so long as no clusters are formed
            for comb in combinations(range(len(self.col_names_all)), 3):
                # By forcing the clusters to be two columns only, we trick syn.smaple() into
                # building exactly the 2dim table
                syn.clusters = Clusters(
                    initial_cluster=[ColumnId(comb[0]), ColumnId(comb[1]), ColumnId(comb[2])],
                    derived_clusters=[],
                )
                df_3col = syn.sample()
                self._parquet_writer(df_3col, f"{self._data_filename(list(df_3col.columns))}.parquet")
                self._build_larger_tables(syn, comb, len(self.col_names_all) - 1)
        # zip up the blob
        self.bzip.zip_blob()

    def _build_larger_tables(self, syn: Synthesizer, comb: tuple[int, ...], maxval: int) -> None:
        comb_id: tuple[ColumnId, ...] = tuple(ColumnId(val) for val in comb)
        next_start = max(comb_id) + 1
        if next_start > maxval:
            return
        for i in range(next_start, maxval + 1):
            build_table = False
            new_comb = comb_id + (ColumnId(i),)
            if len(new_comb) > self.cluster_params.max_cluster_size:
                return
            context = self._clustering_context(new_comb)
            # Test if clustering occurs for a non-targeted set of columns
            clusters = _do_solve(
                context,
                self.cluster_params.default_max_weight,
                self.cluster_params.merge_threshold,
                self.cluster_params.solver_alpha,
            )
            if len(clusters.derived_clusters) == 0:
                build_table = True
            if build_table is False:
                # Test if clustering occurs for each targeted column
                cols_tuple_sorted = self._get_column_names_from_indices(new_comb)
                for target_column in cols_tuple_sorted:
                    features_context = self._features_context(new_comb, cols_tuple_sorted, target_column)
                    clusters = solve_with_features(
                        main_column=features_context.main_column,
                        main_features=features_context.main_features,
                        max_weight=self.cluster_params.ml_max_weight,
                        entropy_1dim=features_context.entropy_1dim,
                        drop_non_features=False,
                    )
                    if len(clusters.derived_clusters) == 0:
                        build_table = True
                        break
            if build_table:
                # There is no clustering here, so we want to generate the table and try larger tables
                # _do_solve has no notion of column indexes. It operates as though the column IDs
                # are in the range 0 to len(new_comb)-1. So we need to convert the column IDs to those
                # in new_comb
                syn.clusters = Clusters(
                    initial_cluster=[ColumnId(c) for c in new_comb],
                    derived_clusters=[],
                )
                df = syn.sample()
                self._parquet_writer(df, f"{self._data_filename(list(df.columns))}.parquet")
                self._build_larger_tables(syn, new_comb, maxval)
                # At this point, we have built any tables made from new_comb or larger, so the
                # tree for new_comb is no longer needed. Delete to save memory.
                del syn.forest._tree_cache[new_comb]

    def _compute_ml_features(self) -> None:
        self.features: dict[str, FeatureSelectionResult] = {}
        for column in self.col_names_all:
            self.features[column] = select_features_ml(
                df=self.df_raw, column=column, classifier_model=None, regressor_model=None, one_hot_X=False
            )

    def _write_features(self) -> None:
        def _round_scores(scores: list[float]) -> list[float]:
            return [round(score, 2) for score in scores]

        features: dict[str, dict[str, Any]] = {}
        for column, feature in self.features.items():
            features[column] = {
                "valid": feature.valid,
                "features": feature.features,
                "k": feature.k,
                "k_features": feature.k_features,
                "cumulative_score": feature.cumulative_score,
                "cumulative_score_std": feature.cumulative_score_std,
                "encoded_scores": feature.encoded_scores,
            }
            # round in order to hide fine details of the data
            features[column]["cumulative_score"] = _round_scores(features[column]["cumulative_score"])
            features[column]["cumulative_score_std"] = _round_scores(features[column]["cumulative_score_std"])
            features[column]["encoded_scores"]["cumulative_score"] = _round_scores(
                features[column]["encoded_scores"]["cumulative_score"]
            )
            features[column]["encoded_scores"]["cumulative_score_std"] = _round_scores(
                features[column]["encoded_scores"]["cumulative_score_std"]
            )
        self._json_writer(features, BlobFiles.FEATURES.value)

    def _write_version(self) -> None:
        self._json_writer(self.write_version, BlobFiles.VERSION.value)

    def _write_measures(self) -> None:
        dependency_matrix_path = self.path_to_blob_dir.joinpath(BlobFiles.DEPENDENCE_MATRIX.value)
        entropy_1dim_path = self.path_to_blob_dir.joinpath(BlobFiles.ENTROPY_1DIM.value)
        np.save(dependency_matrix_path, self.measures.dependency_matrix)
        np.save(entropy_1dim_path, self.measures.entropy_1dim)

    def _parquet_writer(self, column_df: pd.DataFrame, filename: str) -> None:
        file_path = self.path_to_blob_dir.joinpath(filename)
        try:
            column_df.to_parquet(file_path, index=False)
        except Exception as e:
            raise IOError(f"Failed to write to Parquet file {file_path}: {e}") from e

    def _json_writer(self, x: Any, filename: str) -> None:
        file_path = self.path_to_blob_dir.joinpath(filename)
        try:
            with open(file_path, "w", encoding="utf-8") as file:
                json.dump(x, file, indent=4)
        except Exception as e:
            raise IOError(f"Failed to write to file {file_path}: {e}") from e

    def _write_column_names(self) -> None:
        self._json_writer(self.col_names_all, BlobFiles.COLUMNS.value)

    def _write_cluster_params(self) -> None:
        cp = {
            "sample_size": self.cluster_params.sample_size,
            "default_max_weight": self.cluster_params.default_max_weight,
            "ml_max_weight": self.cluster_params.ml_max_weight,
            "merge_threshold": self.cluster_params.merge_threshold,
            "solver_alpha": self.cluster_params.solver_alpha,
            "max_cluster_size": self.cluster_params.max_cluster_size,
        }
        self._json_writer(cp, BlobFiles.CLUSTER_PARAMS.value)

    def _write_bucketization_params(self) -> None:
        bp = {
            "singularity_low_threshold": self.bucketization_params.singularity_low_threshold,
            "range_low_threshold": self.bucketization_params.range_low_threshold,
            "precision_limit_row_fraction": self.bucketization_params.precision_limit_row_fraction,
            "precision_limit_depth_threshold": self.bucketization_params.precision_limit_depth_threshold,
        }
        self._json_writer(bp, BlobFiles.BUCKETIZATION_PARAMS.value)

    def _write_anonymization_params(self) -> None:
        anonymization_params = {
            "low_count_params": {
                "low_threshold": self.anonymization_params.low_count_params.low_threshold,
                "layer_sd": self.anonymization_params.low_count_params.layer_sd,
                "low_mean_gap": self.anonymization_params.low_count_params.low_mean_gap,
            },
            "outlier_count": {
                "lower": self.anonymization_params.outlier_count.lower,
                "upper": self.anonymization_params.outlier_count.upper,
            },
            "top_count": {
                "lower": self.anonymization_params.top_count.lower,
                "upper": self.anonymization_params.top_count.upper,
            },
            "layer_noise_sd": self.anonymization_params.layer_noise_sd,
        }
        self._json_writer(anonymization_params, BlobFiles.ANONYMIZATION_PARAMS.value)


class SyndiffixBlobReader(SyndiffixBlob):
    def __init__(
        self,
        blob_name: str,
        path_to_dir: Optional[Union[str, Path]] = None,
        force: Optional[bool] = False,
        sample_size: Optional[int] = None,
        default_max_weight: Optional[float] = None,
        ml_max_weight: Optional[float] = None,
        merge_threshold: Optional[float] = None,
        solver_alpha: Optional[float] = None,
        cache_df_in_memory: Optional[bool] = True,
    ) -> None:
        super().__init__(blob_name, path_to_dir, force=force)  # Call the base class initializer
        self.read_version: int
        self.columns: list
        self.cache_df_in_memory = cache_df_in_memory
        self.original_column_order: list
        self.catalog = Catalog()
        self.bzip.unzip_blob()
        self._read_cluster_params()
        # We overwrite the cluster parameters if they are passed in. Otherwise they
        # are set as in the blob
        if sample_size is not None:
            self.cluster_params.sample_size = sample_size
        if default_max_weight is not None:
            self.cluster_params.default_max_weight = default_max_weight
        if ml_max_weight is not None:
            self.cluster_params.ml_max_weight = ml_max_weight
        if merge_threshold is not None:
            self.cluster_params.merge_threshold = merge_threshold
        if solver_alpha is not None:
            self.cluster_params.solver_alpha = solver_alpha

        self.measures: DependenceMeasures
        self.context: ClusteringContext
        self.anonymization_params: AnonymizationParams
        self.bucketization_params: BucketizationParams
        self._stitch_record = StitchRecord()

        self._read_version()
        self._read_measures()
        self._read_anonymization_params()
        self._read_bucketization_params()
        self._read_column_names()
        self._read_features()
        self._load_catalog()

    def stitch_record(self) -> List[StitchDict]:
        return self._stitch_record.read()

    def read(self, columns: list[str], target_column: Optional[str] = None) -> pd.DataFrame:
        self._stitch_record.reset()
        columns_copy = columns.copy()
        return self._read(columns_copy, target_column)

    def _read(self, columns: list[str], target_column: Optional[str] = None) -> pd.DataFrame:
        def _check_columns_exist(columns: list[str], col_names_all: list[str]) -> None:
            missing_columns = [col for col in columns if col not in col_names_all]
            if missing_columns:
                raise ValueError(f"Invalid columns: {', '.join(missing_columns)}")

        _check_columns_exist(columns, self.col_names_all)
        # Check for duplicate column names
        if len(set(columns)) != len(columns):
            raise ValueError("Duplicate column names")
        self.original_column_order = columns.copy()
        if target_column is not None and target_column not in self.col_names_all:
            raise ValueError(f"Target column '{target_column}' is not a valid column.")
        columns.sort()
        cols_tuple = tuple(columns)

        if self.catalog.is_valid_key(cols_tuple):
            return self._original_order(self.catalog.read(cols_tuple))

        # Need to stitch!
        comb, cols_tuple_sorted = self._absolute_column_indexes(cols_tuple)
        if target_column is None:
            context = self._clustering_context(comb)
            clusters = _do_solve(
                context,
                self.cluster_params.default_max_weight,
                self.cluster_params.merge_threshold,
                self.cluster_params.solver_alpha,
            )
        else:
            features_context = self._features_context(comb, cols_tuple_sorted, target_column)
            clusters = solve_with_features(
                main_column=features_context.main_column,
                main_features=features_context.main_features,
                max_weight=self.cluster_params.ml_max_weight,
                entropy_1dim=features_context.entropy_1dim,
                drop_non_features=False,
            )
        # The values used in clusters align with the column names in cols_tuple
        self._remap_clusters(clusters, comb)
        columns_left = self._get_column_names_from_indices(clusters.initial_cluster)
        df_left = self._read_catalog(columns_left)
        for _, stitch_ids, other_ids in clusters.derived_clusters:
            stitch_columns = self._get_column_names_from_indices(stitch_ids)
            other_columns = self._get_column_names_from_indices(other_ids)
            columns_right = stitch_columns + other_columns
            df_right = self._read_catalog(columns_right)
            if len(stitch_columns) > 0:
                df_stitch = self._read_catalog(stitch_columns)
                # coerce df_left and df_right to have better quality stitch columns
                df_left = self._stitch(df_left=df_stitch, df_right=df_left, shared=False)
                df_right = self._stitch(df_left=df_stitch, df_right=df_right, shared=False)
            df_left = self._stitch(df_left=df_left, df_right=df_right, shared=True)
        return self._original_order(df_left)

    def _stitch(self, df_left: pd.DataFrame, df_right: pd.DataFrame, shared: bool) -> pd.DataFrame:
        owner = "shared" if shared else "left"
        self._stitch_record.add(list(df_left.columns), list(df_right.columns), owner)
        return stitch(df_left=df_left, df_right=df_right, shared=shared)

    def _remap_clusters(self, clusters: Clusters, comb: tuple[int, ...]) -> None:
        for i, value in enumerate(clusters.initial_cluster):
            clusters.initial_cluster[i] = ColumnId(comb[value])
        for derived_cluster in clusters.derived_clusters:
            for i in range(len(derived_cluster[1])):
                derived_cluster[1][i] = ColumnId(comb[derived_cluster[1][i]])
            for i in range(len(derived_cluster[2])):
                derived_cluster[2][i] = ColumnId(comb[derived_cluster[2][i]])

    def _original_order(self, df: pd.DataFrame) -> pd.DataFrame:
        return df[self.original_column_order]

    def _read_catalog(self, columns: Union[list[str], tuple[str, ...]]) -> pd.DataFrame:
        if isinstance(columns, tuple):
            columns = list(columns)
        columns.sort()
        columns = tuple(columns)
        if not self.catalog.is_valid_key(columns):
            # As a rule, we don't expect this, because we should have built
            # all non-clustered tables. Nevertheless, we should handle it
            return self._recover_missing_df(columns)
        return self.catalog.read(columns)

    def _recover_missing_df(self, columns: tuple[str, ...]) -> pd.DataFrame:
        recover_comb = tuple([ColumnId(self.col_names_all.index(col)) for col in columns])
        stitch_id = max(recover_comb, key=lambda idx: self.measures.entropy_1dim[idx])
        stitch_column = self.col_names_all[stitch_id]
        remaining_columns = [col for col in columns if col != stitch_column]
        midpoint = len(remaining_columns) // 2
        left_columns = remaining_columns[:midpoint] + [stitch_column]
        df_left = self._read_catalog(left_columns)
        right_columns = remaining_columns[midpoint:] + [stitch_column]
        df_right = self._read_catalog(right_columns)
        return self._stitch(df_left=df_left, df_right=df_right, shared=True)

    def _load_catalog(self) -> None:
        for file_path in self.path_to_blob_dir.iterdir():
            if file_path.is_file() and file_path.suffix == ".parquet":
                df = _parquet_reader(file_path)
                columns = list(df.columns)
                columns.sort()
                columns_tuple = tuple(columns)
                if not self.cache_df_in_memory:
                    self.catalog.add(columns_tuple, None, file_path)
                else:
                    self.catalog.add(columns_tuple, df, file_path)

    def _read_features(self) -> None:
        features = self._json_reader(BlobFiles.FEATURES.value)
        self.features: dict[str, FeatureSelectionResult] = {}
        for column, feature in features.items():
            self.features[column] = FeatureSelectionResult(
                valid=feature["valid"],
                features=feature["features"],
                k=feature["k"],
                k_features=feature["k_features"],
                cumulative_score=feature["cumulative_score"],
                cumulative_score_std=feature["cumulative_score_std"],
                encoded_scores=feature["encoded_scores"],
            )

    def _absolute_column_indexes(self, columns: tuple[str, ...]) -> tuple[tuple[int, ...], tuple[str, ...]]:
        indices = [self.col_names_all.index(col) for col in columns]
        paired = list(zip(indices, columns))
        paired_sorted = sorted(paired, key=lambda x: x[0])
        indices_sorted, columns_sorted = zip(*paired_sorted)
        return tuple(indices_sorted), tuple(columns_sorted)

    def _read_version(self) -> None:
        self.read_version = int(self._json_reader(BlobFiles.VERSION.value))

    def _read_measures(self) -> None:
        dependency_matrix_path = self.path_to_blob_dir.joinpath(BlobFiles.DEPENDENCE_MATRIX.value)
        entropy_1dim_path = self.path_to_blob_dir.joinpath(BlobFiles.ENTROPY_1DIM.value)
        self.measures = DependenceMeasures(
            dependency_matrix=np.load(dependency_matrix_path, allow_pickle=True),
            entropy_1dim=np.load(entropy_1dim_path, allow_pickle=True),
        )

    def _json_reader(self, filename: str) -> Any:
        file_path = self.path_to_blob_dir.joinpath(filename)
        try:
            with open(file_path, "r", encoding="utf-8") as file:
                return json.load(file)
        except Exception as e:
            raise IOError(f"Failed to read from file {file_path}: {e}") from e

    def _read_bucketization_params(self) -> None:
        bp = self._json_reader(BlobFiles.BUCKETIZATION_PARAMS.value)
        self.bucketization_params = BucketizationParams(
            singularity_low_threshold=bp["singularity_low_threshold"],
            range_low_threshold=bp["range_low_threshold"],
            precision_limit_row_fraction=bp["precision_limit_row_fraction"],
            precision_limit_depth_threshold=bp["precision_limit_depth_threshold"],
        )

    def _read_column_names(self) -> None:
        self.col_names_all = list(self._json_reader(BlobFiles.COLUMNS.value))

    def _read_cluster_params(self) -> None:
        cp = self._json_reader(BlobFiles.CLUSTER_PARAMS.value)
        self.cluster_params.sample_size = cp["sample_size"]  # type: ignore
        self.cluster_params.default_max_weight = cp["default_max_weight"]  # type: ignore
        self.cluster_params.ml_max_weight = cp["ml_max_weight"]  # type: ignore
        self.cluster_params.merge_threshold = cp["merge_threshold"]  # type: ignore
        self.cluster_params.solver_alpha = cp["solver_alpha"]  # type: ignore

    def _read_anonymization_params(self) -> None:
        ap = self._json_reader(BlobFiles.ANONYMIZATION_PARAMS.value)
        self.anonymization_params = AnonymizationParams(
            salt=b"",
            low_count_params=SuppressionParams(
                low_threshold=ap["low_count_params"]["low_threshold"],
                layer_sd=ap["low_count_params"]["layer_sd"],
                low_mean_gap=ap["low_count_params"]["low_mean_gap"],
            ),
            outlier_count=FlatteningInterval(
                lower=ap["outlier_count"]["lower"],
                upper=ap["outlier_count"]["upper"],
            ),
            top_count=FlatteningInterval(
                lower=ap["top_count"]["lower"],
                upper=ap["top_count"]["upper"],
            ),
            layer_noise_sd=ap["layer_noise_sd"],
        )



import math
import time
from dataclasses import dataclass
from typing import Callable, Optional

import numpy as np
import numpy.typing as npt

from ..common import *
from ..forest import Forest
from ..tree import Branch, Leaf, Node
from .common import *


@dataclass(frozen=True)
class Score:
    score: float
    count: float
    node_xy: Optional[Node]
    node_x: Node
    node_y: Node


@dataclass(frozen=True)
class DependenceMeasure:
    columns: tuple[int, int]
    dependence: float
    measure_time: float


@dataclass(frozen=True)
class DependenceMeasures:
    dependency_matrix: npt.NDArray[np.float64]
    entropy_1dim: Entropy1Dim


def measure_entropy(root: Node) -> float:
    num_rows = root.noisy_count()
    entropy = 0.0

    def entropy_walk(node: Node) -> None:
        nonlocal entropy
        count = node.noisy_count()
        if isinstance(node, Leaf):
            entropy = entropy - (count / num_rows * math.log2(count / num_rows))
        elif isinstance(node, Branch):
            for child_node in node.children.values():
                entropy_walk(child_node)

    entropy_walk(root)
    return entropy


def _is_singularity(node: Node) -> bool:
    return node.actual_intervals[0].is_singularity()


def _singularity_value(node: Node) -> float:
    return node.actual_intervals[0].min


def _get_child(node: Optional[Node], child_id: int) -> Optional[Node]:
    if isinstance(node, Branch):
        return node.children.get(child_id)
    else:
        return None


def _find_child(node: Optional[Node], func: Callable[[int, Node], bool]) -> Optional[Node]:
    if isinstance(node, Branch):
        for key, value in node.children.items():
            if func(key, value):
                return value

    return None


def _get_bit(value: int, index: int) -> int:
    return (value >> index) & 1


def measure_dependence(forest: Forest, col_x: ColumnId, col_y: ColumnId) -> DependenceMeasure:
    if col_x >= col_y:
        raise ValueError("Invalid input.")

    num_rows = forest.data.shape[0]
    range_thresh = forest.bucketization_params.range_low_threshold

    scores: list[Score] = []

    def walk(node_xy: Optional[Node], node_x: Node, node_y: Node) -> None:
        count_x = node_x.noisy_count()
        count_y = node_y.noisy_count()

        if count_x < range_thresh or count_y < range_thresh:
            return

        actual_2dim_count = node_xy.noisy_count() if node_xy else 0.0
        expected_2dim_count = count_x * count_y / num_rows

        score = abs(expected_2dim_count - actual_2dim_count) / max(expected_2dim_count, actual_2dim_count)
        scores.append(Score(score=score, count=expected_2dim_count, node_xy=node_xy, node_x=node_x, node_y=node_y))

        # Walk children.
        # Dim 0 (X) is at bit 1.
        # Dim 1 (Y) is at bit 0.
        if _is_singularity(node_x) and _is_singularity(node_y):
            # Stop walk if both 1-dims are singularities.
            return
        elif _is_singularity(node_x):
            x_singularity = _singularity_value(node_x)

            for id_child_y in range(2):
                child_y = _get_child(node_y, id_child_y)
                if child_y is None:
                    continue

                # Find child that matches on dim Y. It can be on either side of dim X.
                child_xy = _find_child(
                    node_xy,
                    lambda key, value: _get_bit(key, 0) == id_child_y
                    and value.snapped_intervals[0].contains_value(x_singularity),
                )
                walk(child_xy, node_x, child_y)
        elif _is_singularity(node_y):
            y_singularity = _singularity_value(node_y)

            for id_child_x in range(2):
                child_x = _get_child(node_x, id_child_x)
                if child_x is None:
                    continue

                # Find child that matches on dim X. It can be on either side of dim Y.
                child_xy = _find_child(
                    node_xy,
                    lambda key, value: _get_bit(key, 1) == id_child_x
                    and value.snapped_intervals[1].contains_value(y_singularity),
                )
                walk(child_xy, child_x, node_y)
        else:
            for id_child_xy in range(4):
                id_child_x = _get_bit(id_child_xy, 1)
                id_child_y = _get_bit(id_child_xy, 0)

                child_x = _get_child(node_x, id_child_x)
                child_y = _get_child(node_y, id_child_y)

                if child_x is None or child_y is None:
                    continue

                child_xy = _get_child(node_xy, id_child_xy)
                walk(child_xy, child_x, child_y)

    root_xy = forest.get_tree((col_x, col_y))
    root_x = forest.get_tree((col_x,))
    root_y = forest.get_tree((col_y,))

    start_time = time.time()

    walk(root_xy, root_x, root_y)

    total_weighted_score, total_count = 0.0, 0.0

    if len(scores) > 1:
        for i in range(1, len(scores)):
            score = scores[i]
            total_weighted_score += score.score * score.count
            total_count += score.count

    elapsed_time = time.time() - start_time

    return DependenceMeasure(
        columns=(col_x, col_y),
        dependence=total_weighted_score / total_count if total_count > 0 else 0.0,
        measure_time=elapsed_time,
    )


def measure_all(forest: Forest) -> DependenceMeasures:
    num_columns = forest.dimensions
    dependency_matrix = np.full((num_columns, num_columns), 1.0, dtype=float)
    entropy_1dim = np.array([measure_entropy(forest.get_tree((ColumnId(i),))) for i in range(num_columns)], dtype=float)

    for comb in generate_combinations(2, num_columns):
        col_x, col_y = comb
        score = measure_dependence(forest, col_x, col_y)

        dependence = score.dependence
        dependency_matrix[col_x, col_y] = dependence
        dependency_matrix[col_y, col_x] = dependence

    return DependenceMeasures(dependency_matrix=dependency_matrix, entropy_1dim=entropy_1dim)



import os
import sys
import time

import pandas as pd
import psutil

from syndiffix import Synthesizer


# Utility function for loading a CSV file.
def load_csv(path: str) -> pd.DataFrame:
    from pandas.errors import ParserError

    df = pd.read_csv(path, keep_default_na=False, na_values=[""], low_memory=False)

    # Try to infer datetime columns.
    for col in df.columns[df.dtypes == "object"]:
        try:
            df[col] = pd.to_datetime(df[col], format="ISO8601")
        except (ParserError, ValueError):
            pass

    return df


# Simple usage example of the SynDiffix library.
# This script assumes each row belongs to a different protected entity.
# All columns in the input file are processed.


if len(sys.argv) != 3:
    print(f"Usage: py {sys.argv[0]} <input.csv> <output.csv>")
    exit()

input_file = sys.argv[1]
output_file = sys.argv[2]


print(f"Loading data from `{input_file}`...")
input_data = load_csv(input_file)

print(f"Loaded {len(input_data)} rows. Columns:")
for i, (column, dtype) in enumerate(zip(input_data.columns, input_data.dtypes)):
    print(f"{i}: {column} ({dtype})")

start_time = time.time()
process = psutil.Process(os.getpid())
start_memory_usage = process.memory_info().rss

print("\nFitting the synthesizer over the data...")
synthesizer = Synthesizer(input_data)

print("Column clusters:")
print("Initial=", synthesizer.clusters.initial_cluster)
for cluster in synthesizer.clusters.derived_clusters:
    print("Derived=", cluster)

print("\nSampling rows from the synthesizer...")
output_data = synthesizer.sample()

run_time = round(time.time() - start_time)
memory_usage = (process.memory_info().rss - start_memory_usage) // (1024**2)
print(f"Runtime: {run_time} seconds. Memory usage: {memory_usage} MB.")

print(f"\nWriting sampled rows to `{output_file}`...")
output_data.to_csv(output_file, index=False)

print("Done!")



from .blob import SyndiffixBlobBuilder, SyndiffixBlobReader
from .stitcher import stitch
from .synthesizer import Synthesizer

__all__ = [
    "Synthesizer",
    "stitch",
    "SyndiffixBlobBuilder",
    "SyndiffixBlobReader",
]



from dataclasses import dataclass
from typing import Any

import numpy as np
import pandas as pd
from sklearn.compose import ColumnTransformer  # type: ignore
from sklearn.feature_selection import RFECV, VarianceThreshold  # type: ignore
from sklearn.preprocessing import (  # type: ignore
    FunctionTransformer,
    OneHotEncoder,
    OrdinalEncoder,
    RobustScaler,
)
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor  # type: ignore

# ----------------------------------------------------------------
# Feature Selection
# ----------------------------------------------------------------


CATEGORY_THRESHOLD = 15


def _is_categorical(column: pd.Series) -> bool:
    return column.dtype == "object" or column.nunique() <= CATEGORY_THRESHOLD


def _get_feature_types(df: pd.DataFrame) -> tuple[list[str], list[str], list[str]]:
    text_features = []
    continuous_features = []
    categorical_features = []

    for colname in df.columns:
        column = df[colname]
        nunique = column.nunique()
        if nunique <= CATEGORY_THRESHOLD:
            categorical_features.append(colname)
        elif column.dtype == "object":
            text_features.append(colname)
        else:
            continuous_features.append(colname)

    return text_features, continuous_features, categorical_features


def _one_hot_feature_names(encoder: Any, categorical_features: list[str]) -> tuple[list[str], dict[str, str]]:
    cats = [encoder._compute_transformed_categories(i) for i, _ in enumerate(encoder.categories_)]

    feature_names = []
    inverse_names = {}

    for i in range(len(cats)):
        category = categorical_features[i]
        names = [category + "$$" + str(label) for label in cats[i]]
        feature_names.extend(names)

        for name in names:
            inverse_names[name] = category

    return feature_names, inverse_names


def _random_encoder(X: pd.DataFrame) -> pd.DataFrame:
    X = X.copy()

    for column in X.columns:
        column_text = X[column].apply(str)
        unique_values = column_text.unique()
        np.random.shuffle(unique_values)
        mapping = {value: index for index, value in enumerate(unique_values)}
        X[column] = column_text.map(mapping).astype(int)

    return X


def _preprocess(
    df: pd.DataFrame, one_hot_encode: bool = True, variance_threshold: bool = True
) -> tuple[pd.DataFrame, dict[str, str]]:
    # This conversion needed because the blob code works with the original data
    for col in df.select_dtypes(include=["datetime64"]).columns:
        df[col] = df[col].astype(np.int64) // 10**9  # Convert to Unix timestamp
    text_features, continuous_features, categorical_features = _get_feature_types(df)
    ordinal_features = []

    if not one_hot_encode:
        ordinal_features = categorical_features
        categorical_features = []

    preprocessor = ColumnTransformer(
        transformers=[
            ("random_encode", FunctionTransformer(_random_encoder), text_features),
            ("ord", OrdinalEncoder(), ordinal_features),
            ("num", RobustScaler(), continuous_features),
            ("cat", OneHotEncoder(sparse_output=False), categorical_features),
        ],
        sparse_threshold=0,
    )

    result = preprocessor.fit_transform(df)

    inverse_lookup = {name: name for name in text_features + ordinal_features + continuous_features}

    categorical_names: list[str] = []

    if len(categorical_features) > 0:
        categorical_names, categorical_inverse = _one_hot_feature_names(
            preprocessor.named_transformers_["cat"], categorical_features
        )
        inverse_lookup.update(categorical_inverse)

    df_preprocessed = pd.DataFrame(
        result, columns=text_features + ordinal_features + continuous_features + categorical_names
    )

    if not variance_threshold:
        return df_preprocessed, inverse_lookup

    threshold = VarianceThreshold(0.00001)
    threshold.set_output(transform="pandas")
    try:
        df_filtered = threshold.fit_transform(df_preprocessed)
    except ValueError:
        # This should be a rare event. It happens when none of the features have
        # variance above the threshold.
        df_filtered = df_preprocessed

    return df_filtered, inverse_lookup


def _split(df: pd.DataFrame, column: str, one_hot_X: bool) -> tuple[pd.DataFrame, dict[str, str], pd.DataFrame]:
    df = df.dropna(axis=0)
    X, X_inv = _preprocess(df.drop(column, axis=1), one_hot_encode=one_hot_X)
    y, _ = _preprocess(df[[column]], one_hot_encode=False)
    return X, X_inv, y


@dataclass(frozen=True)
class FeatureSelectionResult:
    valid: bool
    features: list[str]
    k: int
    k_features: list[str]
    cumulative_score: list[float]
    cumulative_score_std: list[float]
    encoded_scores: dict


def select_features_ml(
    df: pd.DataFrame,
    column: str,
    classifier_model: Any | None = None,
    regressor_model: Any | None = None,
    one_hot_X: bool = False,
) -> FeatureSelectionResult:
    X, X_inv, y = _split(df, column, one_hot_X)

    if y.shape[0] == 0 or y.shape[1] == 0:
        return FeatureSelectionResult(
            valid=False,
            features=[],
            k=0,
            k_features=[],
            cumulative_score=[],
            cumulative_score_std=[],
            encoded_scores={
                "features": [],
                "k": 0,
                "k_features": [],
                "cumulative_score": [],
                "cumulative_score_std": [],
            },
        )

    if X.shape[1] == 1:
        return FeatureSelectionResult(
            valid=False,
            features=[X.columns[0]],
            k=1,
            k_features=[X.columns[0]],
            cumulative_score=[0.0],
            cumulative_score_std=[0.0],
            encoded_scores={
                "features": [X.columns[0]],
                "k": 1,
                "kFeatures": [X.columns[0]],
                "cumulative_score": [0.0],
                "cumulative_score_std": [0.0],
            },
        )

    if _is_categorical(y[column]):
        estimator = classifier_model or DecisionTreeClassifier(random_state=0)
    else:
        estimator = regressor_model or DecisionTreeRegressor(random_state=0)

    rfecv = RFECV(estimator=estimator)
    rfecv.fit(X, y)

    feature_ranks = rfecv.ranking_
    feature_names = X.columns.tolist()

    sorted_features = sorted(zip(feature_names, feature_ranks), key=lambda x: x[1])

    encoded_k = int(rfecv.n_features_)
    encoded_features = [name for name, _ in sorted_features]
    encoded_scores = rfecv.cv_results_["mean_test_score"].tolist()
    encoded_scores_std = rfecv.cv_results_["std_test_score"].tolist()

    decoded_k = 0
    decoded_features: list[str] = []
    decoded_scores: list[float] = []
    decoded_scores_std: list[float] = []

    for i, feature in enumerate(encoded_features):
        decoded_feature = X_inv[feature]
        if decoded_feature in decoded_features:
            if decoded_features[-1] == decoded_feature:
                decoded_scores[-1] = encoded_scores[i]
                decoded_scores_std[-1] = encoded_scores_std[i]
        else:
            decoded_features.append(decoded_feature)
            decoded_scores.append(encoded_scores[i])
            decoded_scores_std.append(encoded_scores_std[i])

        if i == encoded_k - 1:
            decoded_k = len(decoded_features)

    return FeatureSelectionResult(
        valid=True,
        features=decoded_features,
        k=decoded_k,
        k_features=decoded_features[:decoded_k],
        cumulative_score=decoded_scores,
        cumulative_score_std=decoded_scores_std,
        encoded_scores={
            "features": encoded_features,
            "k": encoded_k,
            "k_features": encoded_features[:encoded_k],
            "cumulative_score": encoded_scores,
            "cumulative_score_std": encoded_scores_std,
        },
    )



from abc import ABC, abstractmethod
from typing import Any, Optional

from ..forest import Forest
from . import features, measures, sampling, solver
from .common import *


def _clustering_context(main_column: Optional[ColumnId], forest: Forest) -> ClusteringContext:
    scores = measures.measure_all(forest)

    total_per_column = [
        sum(scores.dependency_matrix[i, j] for j in range(forest.dimensions) if i != j)
        for i in range(forest.dimensions)
    ]

    return ClusteringContext(
        dependency_matrix=scores.dependency_matrix,
        entropy_1dim=scores.entropy_1dim,
        total_dependence_per_column=total_per_column,
        total_dependence=sum(total_per_column),
        anonymization_params=forest.anonymization_params,
        bucketization_params=forest.bucketization_params,
        rng=forest.unsafe_rng,
        main_column=main_column,
    )


def _resolve_column_id(forest: Forest, col_name: ColumnId | str) -> ColumnId:
    if not isinstance(col_name, str):
        return col_name

    return ColumnId(forest.orig_data.columns.get_loc(col_name))


def _resolve_column_name(forest: Forest, col_id: ColumnId | str) -> str:
    if isinstance(col_id, str):
        return col_id

    return forest.orig_data.columns[col_id]


class ClusteringStrategy(ABC):
    @abstractmethod
    def build_clusters(self, forest: Forest) -> tuple[Clusters, Entropy1Dim]:
        pass


class NoClustering(ClusteringStrategy):
    def build_clusters(self, forest: Forest) -> tuple[Clusters, Entropy1Dim]:
        # Build one patch cluster per column, don't measure entropy.
        assert forest.dimensions > 0
        patch_clusters = Clusters(
            initial_cluster=[ColumnId(0)],
            derived_clusters=[(StitchOwner.SHARED, [], [ColumnId(i)]) for i in range(1, forest.dimensions)],
        )
        return patch_clusters, np.zeros(forest.dimensions, np.float64)


class SingleClustering(ClusteringStrategy):
    def build_clusters(self, forest: Forest) -> tuple[Clusters, Entropy1Dim]:
        # Build and return a cluster that includes everything, don't measure entropy.
        single_cluster = Clusters(initial_cluster=[ColumnId(i) for i in range(forest.dimensions)], derived_clusters=[])
        return single_cluster, np.zeros(forest.dimensions, np.float64)


class DefaultClustering(ClusteringStrategy):
    def __init__(
        self,
        main_column: Optional[ColumnId | str] = None,
        sample_size: int = 1000,
        max_weight: float = 15.0,
        merge_threshold: float = 0.1,
        solver_alpha: float = 1e-2,
    ) -> None:
        self.main_column = main_column
        self.sample_size = sample_size
        self.max_weight = max_weight
        self.merge_threshold = merge_threshold
        self.solver_alpha = solver_alpha

    def build_clusters(self, forest: Forest) -> tuple[Clusters, Entropy1Dim]:
        sampled_forest = (
            sampling.sample_forest(forest, self.sample_size)
            if sampling.should_sample(forest, self.sample_size)
            else forest
        )
        main_column = _resolve_column_id(forest, self.main_column) if self.main_column else None
        clustering_context = _clustering_context(main_column=main_column, forest=sampled_forest)
        clusters = solver.solve(clustering_context, self.max_weight, self.merge_threshold, self.solver_alpha)
        return clusters, clustering_context.entropy_1dim


class MlClustering(ClusteringStrategy):
    def __init__(
        self,
        target_column: ColumnId | str,
        drop_non_features: bool = False,
        max_weight: float = 15.0,
        classifier_model: Any | None = None,
        regressor_model: Any | None = None,
    ) -> None:
        # TODO: Accept ML parameters.
        self.target_column = target_column
        self.drop_non_features = drop_non_features
        self.max_weight = max_weight
        self.classifier_model = classifier_model
        self.regressor_model = regressor_model

    def build_clusters(self, forest: Forest) -> tuple[Clusters, Entropy1Dim]:
        # TODO: Support forest sampling for faster ML feature detection?

        ml_features = features.select_features_ml(
            forest.orig_data,
            _resolve_column_name(forest, self.target_column),
            self.classifier_model,
            self.regressor_model,
        )
        feature_ids = [_resolve_column_id(forest, feature) for feature in ml_features.k_features]

        entropy_1dim = np.array(
            [
                # Solver does not care about entropy of non-features.
                (measures.measure_entropy(forest.get_tree((ColumnId(i),))) if i in feature_ids else 1.0)
                for i in range(forest.dimensions)
            ],
            dtype=float,
        )

        return (
            solver.solve_with_features(
                _resolve_column_id(forest, self.target_column),
                feature_ids,
                self.max_weight,
                entropy_1dim,
                self.drop_non_features,
            ),
            entropy_1dim,
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, replace
from typing import Iterator, NewType, Union

import numpy as np
import numpy.typing as npt

from .anonymizer import hash_strings
from .common import *
from .counters import CountersFactory
from .interval import Interval, Intervals

RowId = NewType("RowId", int)


@dataclass(frozen=True)
class Context:
    combination: Combination
    pid_data: npt.NDArray[Hash]
    data: npt.NDArray[np.float64]
    anonymization_context: AnonymizationContext
    bucketization_params: BucketizationParams
    row_limit: int
    counters_factory: CountersFactory


Subnodes = tuple[Union["Node", None], ...]


class Node(ABC):
    def __init__(
        self, context: Context, subnodes: Subnodes, snapped_intervals: Intervals, actual_intervals: Iterator[Interval]
    ) -> None:
        self.context = context
        self.subnodes = subnodes
        self.snapped_intervals = snapped_intervals
        # These get mutated later, we need to create new Interval instances.
        self.actual_intervals = tuple(interval.copy() for interval in actual_intervals)
        # 0-dim subnodes of 1-dim nodes are not considered stubs.
        self.is_stub = len(subnodes) > 0 and all(subnode is None or subnode.is_stub_subnode() for subnode in subnodes)
        self._stub_subnode_cache: bool | None = None
        self._noisy_count_cache = 0
        self.entity_counter = context.counters_factory.create_entity_counter()

    def dimensions(self) -> int:
        return len(self.context.combination)

    def update_pids(self, row: RowId) -> None:
        self.entity_counter.add(self.context.pid_data[row])

        # We need to recompute cached values each time new contributions arrive.
        self._stub_subnode_cache = None
        self._noisy_count_cache = 0

    def update_actual_intervals(self, row: RowId) -> None:
        values = self.context.data[row]
        for interval, value_index in zip(self.actual_intervals, self.context.combination):
            interval.expand(values[value_index])

    def is_singularity(self) -> bool:
        return all(interval.is_singularity() for interval in self.actual_intervals)

    def is_over_threshold(self, low_threshold: int) -> bool:
        anon_params = self.context.anonymization_context.anonymization_params
        suppression_params = SuppressionParams(
            low_threshold, anon_params.low_count_params.layer_sd, anon_params.low_count_params.low_mean_gap
        )
        return not self.entity_counter.is_low_count(anon_params.salt, suppression_params)

    def is_stub_subnode(self) -> bool:
        if self.is_stub:
            return True
        elif self._stub_subnode_cache is not None:
            return self._stub_subnode_cache  # Return previously cached result.
        else:
            stub_low_threshold = (
                self.context.bucketization_params.singularity_low_threshold
                if self.is_singularity()
                else self.context.bucketization_params.range_low_threshold
            )
            is_stub_subnode = not self.is_over_threshold(stub_low_threshold)
            self._stub_subnode_cache = is_stub_subnode  # Cache result for future tests.
            return is_stub_subnode

    # Helper method for `push_down_1dim_root`.
    # Outliers need special handling. They must:
    #   - be added to existing leaves;
    #   - not change the actual ranges of the nodes.
    @abstractmethod
    def _add_1dim_outlier_row(self, row: RowId) -> None:
        pass

    # Pushes down the root of a 1-dim tree as long as one of the children is low-count.
    @abstractmethod
    def push_down_1dim_root(self) -> Node:
        pass

    @abstractmethod
    def add_row(self, depth: int, row: RowId) -> Node:
        pass

    def bucket_intervals(self) -> Iterator[Interval]:
        for snapped, actual in zip(self.snapped_intervals, self.actual_intervals):
            yield actual if actual.is_singularity() else snapped

    # Returns the noisy count of rows matching the current node.
    def noisy_count(self) -> int:
        if self._noisy_count_cache == 0:
            # Use range midpoints as the labels for seeding the per-bucket noise.
            labels_hash = hash_strings(str(interval.middle()) for interval in self.bucket_intervals())
            bucket_seed = self.context.anonymization_context.bucket_seed ^ labels_hash
            anon_context = replace(self.context.anonymization_context, bucket_seed=bucket_seed)

            row_counter = self.context.counters_factory.create_row_counter()
            for row in self._matching_rows():
                row_counter.add(self.context.pid_data[row])

            min_count = anon_context.anonymization_params.low_count_params.low_threshold
            self._noisy_count_cache = max(row_counter.noisy_count(anon_context), min_count)

        return self._noisy_count_cache

    # Helper method for `noisy_count`. Iterates over all of the rows matching the current node.
    @abstractmethod
    def _matching_rows(self) -> Iterator[RowId]:
        pass


class Leaf(Node):
    def __init__(self, context: Context, subnodes: Subnodes, snapped_intervals: Intervals, initial_row: RowId):
        initial_values = get_items_combination(context.combination, context.data[initial_row])
        actual_intervals = (Interval(value, value) for value in initial_values)
        super().__init__(context, subnodes, snapped_intervals, actual_intervals)
        self.rows = [initial_row]

        self.update_pids(initial_row)
        self.update_actual_intervals(initial_row)

    def _should_split(self, depth: int) -> bool:
        low_threshold = self.context.anonymization_context.anonymization_params.low_count_params.low_threshold
        depth_threshold = self.context.bucketization_params.precision_limit_depth_threshold

        return (
            # `row_limit` is 0 for trees higher-dimensional than 1, no need to check dimensions here.
            (depth <= depth_threshold or len(self.rows) >= self.context.row_limit)
            and not self.is_stub
            and not self.is_singularity()
            and self.is_over_threshold(low_threshold)
        )

    def add_row(self, depth: int, row: RowId) -> Node:
        self.update_pids(row)
        self.update_actual_intervals(row)
        self.rows.append(row)

        if self._should_split(depth):
            # Convert current leaf node into a new branch node and insert previously gathered rows down the tree.
            branch: Node = Branch(self)
            for row in self.rows:
                branch = branch.add_row(depth, row)
            return branch
        else:
            return self

    def _add_1dim_outlier_row(self, row: RowId) -> None:
        self.update_pids(row)
        self.rows.append(row)

    def push_down_1dim_root(self) -> Node:
        return self

    def _matching_rows(self) -> Iterator[RowId]:
        yield from self.rows


class Branch(Node):
    def __init__(self, leaf: Leaf):
        super().__init__(leaf.context, leaf.subnodes, leaf.snapped_intervals, iter(leaf.actual_intervals))
        self.children: dict[int, Node] = dict()

    # Each dimension corresponds to a bit in index, with 0 for the lower interval half and
    # 1 for the the upper interval half.
    def _find_child_index(self, row: RowId) -> int:
        values = self.context.data[row]
        child_index = 0
        for value_index, interval in zip(self.context.combination, self.snapped_intervals):
            child_index = (child_index << 1) | interval.half_index(values[value_index])
        return child_index

    @staticmethod
    def _remove_dimension_from_index(position: int, index: int) -> int:
        lower_mask = (1 << position) - 1
        upper_mask = ~((1 << (position + 1)) - 1)
        # Remove missing position bit from index.
        return ((index & upper_mask) >> 1) | (index & lower_mask)

    def _create_child_leaf(self, child_index: int, initial_row: RowId) -> Leaf:
        # Create child's intervals by halfing parent's intervals, using the corresponding bit
        # in the index to select the correct half.
        dimensions = self.dimensions()
        snapped_intervals = tuple(
            interval.half((child_index >> (dimensions - dim_index)) & 1)
            for dim_index, interval in enumerate(self.snapped_intervals, 1)
        )

        # Set child's subnodes to the matching-interval children of the parent's subnodes.
        subnodes = tuple(
            (
                subnode.children.get(Branch._remove_dimension_from_index(dim_index, child_index))
                if isinstance(subnode, Branch)
                else None
            )
            for dim_index, subnode in enumerate(self.subnodes)
        )

        return Leaf(self.context, subnodes, snapped_intervals, initial_row)

    def add_row(self, depth: int, row: RowId) -> Node:
        child_index = self._find_child_index(row)
        child = self.children.get(child_index)
        self.children[child_index] = (
            self._create_child_leaf(child_index, row) if child is None else child.add_row(depth + 1, row)
        )
        self.update_pids(row)
        self.update_actual_intervals(row)
        return self

    def _add_1dim_outlier_row(self, row: RowId) -> None:
        self.update_pids(row)
        child_index = next(iter(self.children)) if len(self.children) == 1 else self._find_child_index(row)
        self.children[child_index]._add_1dim_outlier_row(row)

    # Returns the low-count rows, if any, from the specified child leaf of the current 1-dim branch.
    def _get_low_count_rows_in_child(self, child_index: int) -> list[RowId] | None:
        child = self.children.get(child_index)
        if child is None:
            return []
        elif isinstance(child, Leaf):
            low_threshold = self.context.anonymization_context.anonymization_params.low_count_params.low_threshold
            return None if child.is_over_threshold(low_threshold) else child.rows
        else:
            return None

    def push_down_1dim_root(self) -> Node:
        left_lc_rows = self._get_low_count_rows_in_child(0)
        right_lc_rows = self._get_low_count_rows_in_child(1)
        if left_lc_rows is None and right_lc_rows is not None:
            new_root = self.children[0].push_down_1dim_root()
            for row in right_lc_rows:
                new_root._add_1dim_outlier_row(row)
            return new_root
        elif left_lc_rows is not None and right_lc_rows is None:
            new_root = self.children[1].push_down_1dim_root()
            for row in left_lc_rows:
                new_root._add_1dim_outlier_row(row)
            return new_root
        else:
            return self

    def _matching_rows(self) -> Iterator[RowId]:
        for child in self.children.values():
            yield from child._matching_rows()



import hashlib
import math
import operator
import sys
from collections import Counter
from dataclasses import dataclass, field, replace
from functools import reduce
from typing import Iterator, cast

from .common import *


@dataclass
class PidContributions:
    value_counts: Counter[Hash] = field(default_factory=Counter)
    unaccounted_for: int = 0


@dataclass(frozen=True)
class CountResult:
    anonymized_count: int
    noise_sd: float


# ----------------------------------------------------------------
# Noise
# ----------------------------------------------------------------

# The noise seeds are hash values.
# From each seed we generate a single random value, with either a uniform or a normal distribution.
# Any decent hash function should produce values that are uniformly distributed over the output space.
# Hence, we only need to limit the seed to the requested interval to get a uniform random integer.
# To get a normal random float, we use the Box-Muller method on two uniformly distributed integers.


def _random_uniform(interval: FlatteningInterval, seed: Hash) -> int:
    # While using modulo to bound values produces biased output, we are using very small ranges
    # (typically less than 10), for which the bias is insignificant.
    return int(seed) % (interval.upper - interval.lower + 1) + interval.lower


def _random_normal(sd: float, seed: Hash) -> float:
    u1 = (int(seed) & 0x7FFFFFFF) / 0x7FFFFFFF
    u1 = max(u1, sys.float_info.epsilon)
    u2 = ((int(seed) >> 32) & 0x7FFFFFFF) / 0x7FFFFFFF
    normal = math.sqrt(-2.0 * math.log(u1)) * math.sin(2.0 * math.pi * u2)
    return sd * normal


def _crypto_hash_salted_seed(salt: bytes, seed: Hash) -> Hash:
    hash = hashlib.sha256(salt + seed.tobytes()).digest()
    return Hash(int.from_bytes(hash[:8], "little"))


def _hash_bytes(b: bytes) -> Hash:
    hash = hashlib.blake2b(b, digest_size=8).digest()
    return Hash(int.from_bytes(hash, "little"))


def _hash_string(s: str) -> Hash:
    return _hash_bytes(s.encode())


def _hash_int(i: int) -> Hash:
    return _hash_bytes(i.to_bytes(8, "little"))


def _mix_seed(step_name: str, seed: Hash) -> Hash:
    return _hash_string(step_name) ^ seed


def _generate_noise(salt: bytes, step_name: str, sd: float, noise_layers: tuple[Hash, ...]) -> float:
    noise = 0.0
    for layer_seed in noise_layers:
        noise += _random_normal(sd, _mix_seed(step_name, _crypto_hash_salted_seed(salt, layer_seed)))
    return noise


def _compact_flattening_intervals(
    outlier_count: FlatteningInterval, top_count: FlatteningInterval, total_count: int
) -> tuple[FlatteningInterval, FlatteningInterval] | None:
    if total_count < outlier_count.lower + top_count.lower:
        return None

    total_adjustment = outlier_count.upper + top_count.upper - total_count

    outlier_upper = outlier_count.upper
    top_upper = top_count.upper

    if total_adjustment > 0:
        # NOTE: At this point we know `0 < total_adjustment <= outlier_range + top_range` (*):
        #       `total_adjustment = outlier_count.upper + top_count.upper - total_count
        #                        <= outlier_count.upper + top_count.upper - outlier_count.lower - top_count.lower`.
        outlier_range = outlier_count.upper - outlier_count.lower
        top_range = top_count.upper - top_count.lower
        # `top_adjustment` will be half of `total_adjustment` rounded up, so it takes priority as it should.
        outlier_adjustment = total_adjustment // 2
        top_adjustment = total_adjustment - outlier_adjustment

        # Adjust, depending on how the adjustments "fit" in the ranges.
        match (outlier_range >= outlier_adjustment, top_range >= top_adjustment):
            case (True, True):
                # Both ranges are compacted at same rate.
                outlier_upper -= outlier_adjustment
                top_upper -= top_adjustment
            case (False, True):
                # `outlier_count` is compacted as much as possible by `outlier_range`,
                # `top_count` takes the surplus adjustment.
                outlier_upper = outlier_count.lower
                top_upper -= total_adjustment - outlier_range
            case (True, False):
                # Vice versa.
                outlier_upper -= total_adjustment - top_range
                top_upper = top_count.lower
            case (False, False):
                # Not possible. Otherwise:
                # `outlier_range + top_range < outlier_adjustment + top_adjustment = total_adjustment`,
                # but we knew the opposite was true in (*) above.
                raise RuntimeError("Impossible interval compacting.")

    return replace(outlier_count, upper=outlier_upper), replace(top_count, upper=top_upper)


@dataclass(frozen=True)
class _PidCount:
    flattened_count: float
    flattening: float
    noise_sd: float
    noise: float


def _flatten_contributions(pid_contributions: PidContributions, context: AnonymizationContext) -> _PidCount | None:
    total_count = len(pid_contributions.value_counts)
    anon_params = context.anonymization_params

    flattening_intervals = _compact_flattening_intervals(anon_params.outlier_count, anon_params.top_count, total_count)
    if flattening_intervals is None:
        return None  # Insufficient values for a sensible flattening.

    outlier_interval, top_interval = flattening_intervals

    # Sort contributions by amount and PID value (to ensure consistency in case of equal contributions).
    sorted_value_counts = sorted(
        pid_contributions.value_counts.items(),
        reverse=True,
        key=operator.itemgetter(1, 0),
    )

    flat_seed = seed_from_pid_set(pid for pid, _ in sorted_value_counts[: outlier_interval.upper + top_interval.upper])
    flat_seed = _crypto_hash_salted_seed(anon_params.salt, flat_seed)
    outlier_count = _random_uniform(outlier_interval, _mix_seed("outlier", flat_seed))
    top_count = _random_uniform(top_interval, _mix_seed("top", flat_seed))

    top_group_sum = sum(
        contribution for _, contribution in sorted_value_counts[outlier_count : (outlier_count + top_count)]
    )
    top_group_average = top_group_sum / top_count

    flattening = sum(
        max(contribution - top_group_average, 0) for _, contribution in sorted_value_counts[:outlier_count]
    )

    real_sum = pid_contributions.value_counts.total()
    flattened_unaccounted_for = max(pid_contributions.unaccounted_for - flattening, 0)
    flattened_sum = real_sum - flattening
    flattened_avg = flattened_sum / total_count

    noise_scale = max(flattened_avg, 0.5 * top_group_average)
    noise_sd = anon_params.layer_noise_sd * noise_scale

    pid_seed = seed_from_pid_set(pid_contributions.value_counts)
    noise = _generate_noise(anon_params.salt, "noise", noise_sd, (context.bucket_seed, pid_seed))

    return _PidCount(flattened_sum + flattened_unaccounted_for, flattening, noise_sd, noise)


def _anonymized_sum(pid_counts: Iterable[_PidCount]) -> tuple[float, float]:
    # We might end up with multiple different flattened counts that have the same amount of flattening.
    # This could be the result of some PID values being null for one of the PIDs, while there were still
    # overall enough PIDs to produce a flattened count.
    # In these cases, we want to use the largest flattened count to minimize unnecessary flattening.
    flattening = max(pid_counts, key=lambda count: (count.flattening, count.flattened_count))

    # For determinism, resolve draws using the maximum absolute noise value.
    noise = max(pid_counts, key=lambda count: (count.noise_sd, abs(count.noise)))

    return flattening.flattened_count + noise.noise, noise.noise_sd


MONEY_ROUND_MIN = 1e-10
MONEY_ROUND_DELTA = MONEY_ROUND_MIN / 100.0


# Works with `value` between 1.0 and 10.0.
def _money_round_internal(value: float) -> float:
    if 1.0 <= value < 1.5:
        return 1.0
    elif 1.5 <= value < 3.5:
        return 2.0
    elif 3.5 <= value < 7.5:
        return 5.0
    else:
        return 10.0


def _money_round(value: float) -> float:
    if 0.0 <= value < MONEY_ROUND_MIN:
        return 0.0

    tens = 10.0 ** math.floor(math.log10(value))
    return tens * _money_round_internal(value / tens)


def _money_round_noise(noise_sd: float) -> float:
    if noise_sd == 0.0:
        return 0.0

    rounding_resolution = _money_round(0.05 * noise_sd)
    return math.ceil(noise_sd / rounding_resolution) * rounding_resolution


# ----------------------------------------------------------------
# Public API
# ----------------------------------------------------------------


def hash_strings(strings: Iterator[str]) -> Hash:
    return seed_from_pid_set(_hash_string(string) for string in set(strings))


def hash_pid(pid: object) -> Hash:
    if not pid:
        return Hash(0)
    elif isinstance(pid, int):
        return _hash_int(cast(int, pid))
    elif isinstance(pid, str):
        return _hash_string(cast(str, pid))
    else:
        raise NotImplementedError("Unsupported PID type!")


def seed_from_pid_set(pid_set: Iterable[Hash]) -> Hash:
    return reduce(operator.xor, pid_set, Hash(0))


# Returns whether any of the PID value sets has a low count.
def is_low_count(salt: bytes, params: SuppressionParams, pid_trackers: list[tuple[int, Hash]]) -> bool:
    assert len(pid_trackers) > 0

    for count, seed in pid_trackers:
        if count < params.low_threshold:
            return True

        noise = _generate_noise(salt, "suppress", params.layer_sd, (seed,))

        # `low_mean_gap` is the number of standard deviations between `low_threshold` and desired mean.
        mean = params.low_mean_gap * params.layer_sd + params.low_threshold

        if count < noise + mean:
            return True

    return False


def count_multiple_contributions(
    context: AnonymizationContext, contributions_list: list[PidContributions]
) -> CountResult | None:
    assert len(contributions_list) > 0

    flattened_contributions = [_flatten_contributions(contributions, context) for contributions in contributions_list]

    # If any of the PIDs had insufficient data to produce a sensible flattening, we have to abort anonymization.
    if not all(flattened_contributions):
        return None

    value, noise_sd = _anonymized_sum(cast(list[_PidCount], flattened_contributions))

    return CountResult(round(value), _money_round_noise(noise_sd))


def count_single_contributions(context: AnonymizationContext, count: int, seed: Hash) -> int:
    params = context.anonymization_params
    noise = _generate_noise(params.salt, "noise", params.layer_noise_sd, (context.bucket_seed, seed))
    return round(count + noise)


def noisy_row_limit(salt: bytes, seed: Hash, row_count: int, row_fraction: int) -> int:
    real_row_limit = row_count // row_fraction

    # Select an integer between plus and minus 5% of `real_row_limit`.
    noise_range = real_row_limit // 20
    noise_seed = _mix_seed("precision_limit", _crypto_hash_salted_seed(salt, seed))
    noise = _random_uniform(FlatteningInterval(-noise_range, noise_range), noise_seed)

    return real_row_limit + noise



from __future__ import annotations

import math
from dataclasses import dataclass


@dataclass
class Interval:
    # Notice: this won't work for integer values bigger than 2^53 because of floating point limitations.
    min: float
    max: float

    def __post_init__(self) -> None:
        assert self.min <= self.max

    def size(self) -> float:
        return self.max - self.min

    def is_singularity(self) -> bool:
        return self.min == self.max

    def middle(self) -> float:
        return self.min if self.is_singularity() else (self.min + self.max) / 2.0

    # Returns 0 if value is in interval's lower half, 1 if value is in interval's upper half.
    def half_index(self, value: float) -> int:
        return 0 if self.is_singularity() or value < self.middle() else 1

    def lower_half(self) -> Interval:
        return Interval(self.min, self.middle())

    def upper_half(self) -> Interval:
        return Interval(self.middle(), self.max)

    def half(self, index: int) -> Interval:
        match index:
            case 0:
                return self.lower_half()
            case 1:
                return self.upper_half()
            case _:
                raise ValueError("Invalid index for interval half!")

    def contains_interval(self, interval: Interval) -> bool:
        return interval.min >= self.min and interval.max <= self.max

    def contains_value(self, value: float) -> bool:
        return value == self.min or (value > self.min and value < self.max)

    def overlaps(self, interval: Interval) -> bool:
        return interval.min < self.max and interval.max > self.min

    def expand(self, value: float) -> None:
        if value > self.max:
            self.max = value
        elif value < self.min:
            self.min = value

    def copy(self) -> Interval:
        return Interval(self.min, self.max)


Intervals = tuple[Interval, ...]


def _next_pow2(x: float) -> float:
    assert x > 0.0
    return 2.0 ** math.ceil(math.log2(x))


def _floor_by(value: float, amount: float) -> float:
    return math.floor(value / amount) * amount


def snap_interval(interval: Interval) -> Interval:
    snapped_size = _next_pow2(interval.size()) if interval.size() > 0.0 else 1.0
    aligned_min = _floor_by(interval.min, snapped_size / 2.0)
    if aligned_min + snapped_size < interval.max:
        # This snapped interval doesn't fit, so we need to increase it.
        return snap_interval(Interval(aligned_min, interval.max))
    else:
        return Interval(aligned_min, aligned_min + snapped_size)


def get_null_mapping(interval: Interval) -> float:
    if interval.max > 0:
        return 2 * interval.max
    elif interval.min < 0:
        return 2 * interval.min
    else:
        return 1.0



from abc import ABC, abstractmethod
from bisect import bisect_left
from itertools import islice
from os.path import commonprefix
from random import Random
from typing import Generator, Iterable, Literal, Optional, Set, cast

import numpy as np
import pandas as pd
from pandas.api.types import (
    is_bool_dtype,
    is_datetime64_dtype,
    is_float_dtype,
    is_integer_dtype,
    is_string_dtype,
)
from sklearn.preprocessing import MinMaxScaler

from .bucket import Buckets
from .common import ColumnType, Value
from .interval import Interval, Intervals
from .tree import Branch, Leaf, Node

MICRODATA_SYN_VALUE: Literal[0] = 0
MICRODATA_FLOAT_VALUE: Literal[1] = 1

MicrodataValue = tuple[Value, float]
MicrodataRow = list[MicrodataValue]


TIMESTAMP_REFERENCE = pd.Timestamp("1800-01-01T00:00:00")


class DataConvertor(ABC):
    def __init__(self) -> None:
        self.scaler: Optional[MinMaxScaler] = None

    @abstractmethod
    def column_type(self) -> ColumnType:
        pass

    @abstractmethod
    def to_float(self, value: Value) -> float:
        pass

    @abstractmethod
    def from_interval(self, interval: Interval, rng: Random) -> MicrodataValue:
        pass

    def analyze_tree(self, root: Node) -> None:
        pass


class BooleanConvertor(DataConvertor):
    def __init__(self) -> None:
        super().__init__()

    def column_type(self) -> ColumnType:
        return ColumnType.BOOLEAN

    def to_float(self, value: Value) -> float:
        assert isinstance(value, bool) or isinstance(value, np.bool_)
        return 1.0 if value else 0.0

    def from_interval(self, interval: Interval, rng: Random) -> MicrodataValue:
        value = _generate_float(interval, rng) >= 0.5
        return (value, 1.0 if value else 0.0)


class RealConvertor(DataConvertor):
    def __init__(self, values: Iterable[Value]) -> None:
        super().__init__()
        # Fit up to 0.9999 so that the max bucket range is 0-1
        self.scaler = MinMaxScaler(feature_range=(0, 0.9999))
        # This value-neutral fitting is only for passing unit tests.
        self.scaler.fit(np.array([[0.0], [0.9999]]))
        self.round_precision = _get_round_precision(values)

    def column_type(self) -> ColumnType:
        return ColumnType.REAL

    def to_float(self, value: Value) -> float:
        assert isinstance(value, float) or isinstance(value, np.floating)
        return float(value)

    def from_interval(self, interval: Interval, rng: Random) -> MicrodataValue:
        value = _generate_float(interval, rng)
        value = _inverse_normalize_value(value, self.scaler)
        value = round(value, self.round_precision)
        return (value, value)


class IntegerConvertor(DataConvertor):
    def __init__(self) -> None:
        super().__init__()
        # Fit up to 0.9999 so that the max bucket range is 0-1
        self.scaler = MinMaxScaler(feature_range=(0, 0.9999))
        # This value-neutral fitting is only for passing unit tests.
        self.scaler.fit(np.array([[0.0], [0.9999]]))

    def column_type(self) -> ColumnType:
        return ColumnType.INTEGER

    def to_float(self, value: Value) -> float:
        assert isinstance(value, int) or isinstance(value, np.integer)
        return float(value)

    def from_interval(self, interval: Interval, rng: Random) -> MicrodataValue:
        value = _generate_float(interval, rng)
        value = _inverse_normalize_value(value, self.scaler)
        value = round(value)
        return (value, float(value))


class TimestampConvertor(DataConvertor):
    def __init__(self) -> None:
        super().__init__()
        # Fit up to 0.9999 so that the max bucket range is 0-1
        self.scaler = MinMaxScaler(feature_range=(0, 0.9999))
        # This value-neutral fitting is only for passing unit tests.
        self.scaler.fit(np.array([[0.0], [0.9999]]))

    def column_type(self) -> ColumnType:
        return ColumnType.TIMESTAMP

    def to_float(self, value: Value) -> float:
        assert isinstance(value, pd.Timestamp)
        # converting date time into second timestamp, counting from reference.
        return float((value - TIMESTAMP_REFERENCE) / pd.Timedelta(1, "s"))

    def from_interval(self, interval: Interval, rng: Random) -> MicrodataValue:
        value = _generate_float(interval, rng)
        value = _inverse_normalize_value(value, self.scaler)
        datetime = TIMESTAMP_REFERENCE + np.timedelta64(int(value), "s")
        return (datetime, value)


class StringConvertor(DataConvertor):
    def __init__(self, values: Iterable[Value]) -> None:
        super().__init__()
        unique_values = set(v for v in values if not pd.isna(v))
        for value in unique_values:
            if not isinstance(value, str):
                raise TypeError(f"Not a `str` object in a string dtype column: {value}.")
        self.value_map = sorted(cast(Set[str], unique_values))
        self.safe_values: Set[int] = set()

    def column_type(self) -> ColumnType:
        return ColumnType.STRING

    def to_float(self, value: Value) -> float:
        index = bisect_left(self.value_map, cast(str, value))
        assert index >= 0 and index < len(self.value_map)
        return float(index)

    def from_interval(self, interval: Interval, rng: Random) -> MicrodataValue:
        if interval.is_singularity():
            return (self.value_map[int(interval.min)], interval.min)
        else:
            return self._map_interval(interval, rng)

    def _map_interval(self, interval: Interval, rng: Random) -> MicrodataValue:
        # If a randomly selected value from the interval is not safe, finds a
        # common prefix of the strings encoded as interval boundaries and
        # appends "*" and a random number to ensure that the count of distinct
        # values approximates that in the original data.
        min_value = int(interval.min)
        # max_value is inclusive
        max_value = min(int(interval.max) - 1, len(self.value_map) - 1)
        # The latter term in the above line can 0 (not sure why TODO: check)
        max_value = max(min_value, max_value)
        value = rng.randint(min_value, max_value)
        if value in self.safe_values:
            return (self.value_map[value], float(value))
        else:
            return (
                commonprefix([self.value_map[min_value], self.value_map[max_value]]) + "*" + str(value),
                float(value),
            )

    def analyze_tree(self, root: Node) -> None:
        def analyze_tree_walk(node: Node) -> None:
            if isinstance(node, Leaf):
                low_threshold = node.context.anonymization_context.anonymization_params.low_count_params.low_threshold
                if node.is_singularity() and node.is_over_threshold(low_threshold):
                    self.safe_values.add(int(node.actual_intervals[0].min))
            elif isinstance(node, Branch):
                for child_node in node.children.values():
                    analyze_tree_walk(child_node)

        analyze_tree_walk(root)


def _generate_float(interval: Interval, rng: Random) -> float:
    return rng.uniform(interval.min, interval.max)


def _get_round_precision(values: Iterable[Value]) -> int:
    max_precision = 0
    for value in values:
        assert isinstance(value, float) or isinstance(value, np.floating)
        value_str = str(value)
        if "." in value_str:
            decimal_part = value_str.split(".")[1]
            precision = len(decimal_part)
        else:
            precision = 0
        if precision > max_precision:
            max_precision = precision
    return max_precision


def _generate(interval: Interval, convertor: DataConvertor, null_mapping: float, rng: Random) -> MicrodataValue:
    return convertor.from_interval(interval, rng) if interval.min != null_mapping else (None, null_mapping)


def _microdata_row_generator(
    intervals: Intervals, convertors: list[DataConvertor], null_mappings: list[float], rng: Random
) -> Generator[MicrodataRow, None, None]:
    assert len(intervals) == len(convertors)
    assert len(intervals) == len(null_mappings)
    while True:
        yield [_generate(i, c, nm, rng) for i, c, nm in zip(intervals, convertors, null_mappings)]


def get_convertor(df: pd.DataFrame, column: str) -> DataConvertor:
    dtype = df.dtypes[column]
    if is_integer_dtype(dtype):
        return IntegerConvertor()
    elif is_float_dtype(dtype):
        return RealConvertor(df[column])
    elif is_bool_dtype(dtype):
        return BooleanConvertor()
    elif is_datetime64_dtype(dtype):
        return TimestampConvertor()
    elif is_string_dtype(dtype):
        # Note above is `True` for `object` dtype, but `StringConvertor` will assert values are `str`.
        return StringConvertor(df[column])
    else:
        raise TypeError(f"Dtype {dtype} is not supported.")


def _inverse_normalize_value(value: float, scaler: MinMaxScaler) -> float:
    # Inverse of normalize, but for one value at a time
    value_array = np.array([[value]])
    inverse_transformed_array = scaler.inverse_transform(value_array)
    inverse_transformed_value = inverse_transformed_array[0, 0]
    return float(inverse_transformed_value)


def _normalize(values: pd.Series, scaler: Optional[MinMaxScaler]) -> pd.Series:
    if scaler is None:
        # Convertors that don't need normalization
        return values

    # MinMax normalize values, while retaining the NaN values
    values_array = values.to_numpy()
    nan_indices = np.isnan(values_array)
    if nan_indices.all():
        return values
    median_value = np.nanmedian(values_array)
    values_array[nan_indices] = median_value
    values_reshaped = values_array.reshape(-1, 1)
    normalized_values = scaler.fit_transform(values_reshaped).flatten()
    normalized_values[nan_indices] = np.nan
    return pd.Series(normalized_values, index=values.index)


def _apply_convertor(value: Value, convertor: DataConvertor) -> float:
    if pd.isna(value):
        return np.nan
    else:
        return convertor.to_float(value)


def apply_convertors(convertors: list[DataConvertor], raw_data: pd.DataFrame) -> pd.DataFrame:
    converted_columns = [
        raw_data[column].apply(_apply_convertor, convertor=convertor)
        for column, convertor in zip(raw_data.columns, convertors)
    ]

    possibly_normalized_columns = [
        _normalize(column, convertor.scaler) for column, convertor in zip(converted_columns, convertors)
    ]

    return pd.DataFrame(dict(zip(raw_data.columns, possibly_normalized_columns)), copy=False)


def generate_microdata(
    buckets: Buckets, convertors: list[DataConvertor], null_mappings: list[float], rng: Random
) -> list[MicrodataRow]:
    microdata_rows: list[MicrodataRow] = []
    for bucket in buckets:
        microdata_rows.extend(
            islice(_microdata_row_generator(bucket.intervals, convertors, null_mappings, rng), bucket.count)
        )

    return microdata_rows



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, replace
from random import Random
from typing import Any, Callable, Sequence

from ..common import *
from ..interval import Interval
from ..microdata import MICRODATA_FLOAT_VALUE, MicrodataRow
from .common import *


@dataclass
class ColumnLocation:
    source_row: StitchOwner
    column_id: ColumnId
    left_index: ColumnIndex
    right_index: ColumnIndex


@dataclass(frozen=True)
class StitchContext:
    rng: Random
    stitch_owner: StitchOwner
    all_columns: list[ColumnLocation]
    entropy_1dim: Entropy1Dim
    stitch_max_values: list[float]
    stitch_is_integral: list[bool]
    left_stitch_indexes: list[ColumnIndex]
    right_stitch_indexes: list[ColumnIndex]
    result_rows: list[MicrodataRow]

    @property
    def num_stitch_columns(self) -> int:
        return len(self.stitch_is_integral)


@dataclass(frozen=True)
class StitchState:
    depth: int
    stitch_intervals: list[Interval]
    next_sort_column: int
    currently_sorted_by: int
    remaining_sort_attempts: int
    context: StitchContext

    def replace(self, **kwargs: Any) -> StitchState:
        return replace(self, **kwargs)


def _align_length(rng: Random, length: int, microtable: list[MicrodataRow]) -> list[MicrodataRow]:
    curr_length = len(microtable)
    if length == curr_length:
        return microtable
    elif length < curr_length:
        return microtable[:length]
    else:
        microtable_copy = microtable.copy()
        microtable_copy.extend([microtable[rng.randint(0, curr_length - 1)] for _ in range(length - curr_length)])
        return microtable_copy


def _find_indexes(subset: Sequence[ColumnId], superset: Sequence[ColumnId]) -> list[ColumnIndex]:
    assert len(subset) <= len(superset)
    return [ColumnIndex(superset.index(c)) for c in subset]


def _locate_columns(left_combination: Combination, right_combination: Combination) -> list[ColumnLocation]:
    column_locations: dict[ColumnId, ColumnLocation] = defaultdict(
        lambda: ColumnLocation(StitchOwner.LEFT, ColumnId(-1), ColumnIndex(-1), ColumnIndex(-1))
    )

    for column_index, column_id in enumerate(left_combination):
        location = column_locations[column_id]
        location.column_id = column_id
        location.left_index = ColumnIndex(column_index)
        column_locations[column_id] = location

    for column_index, column_id in enumerate(right_combination):
        location = column_locations[column_id]
        location.column_id = column_id
        location.right_index = ColumnIndex(column_index)
        location.source_row = StitchOwner.RIGHT if location.left_index < 0 else StitchOwner.SHARED
        column_locations[column_id] = location

    return sorted(column_locations.values(), key=lambda x: x.column_id)


def _binary_search(rows: list[MicrodataRow], column: ColumnIndex, target: float, start: int, end_exclusive: int) -> int:
    if start >= end_exclusive:
        return -1

    mid = (start + end_exclusive) // 2

    if rows[mid][column][MICRODATA_FLOAT_VALUE] >= target:
        if mid == 0 or rows[mid - 1][column][MICRODATA_FLOAT_VALUE] < target:
            return mid
        else:
            return _binary_search(rows, column, target, start, mid)
    else:
        return _binary_search(rows, column, target, mid + 1, end_exclusive)


DUMMY_VALUE = (0.0, 0.0)


def _merge_row(
    all_columns: list[ColumnLocation], pick_shared_left: bool, left_row: MicrodataRow, right_row: MicrodataRow
) -> MicrodataRow:
    merged_row: MicrodataRow = [DUMMY_VALUE] * len(all_columns)

    for j, col in enumerate(all_columns):
        if col.source_row == StitchOwner.LEFT:
            merged_row[j] = left_row[col.left_index]
        elif col.source_row == StitchOwner.RIGHT:
            merged_row[j] = right_row[col.right_index]
        elif pick_shared_left:
            merged_row[j] = left_row[col.left_index]
        else:
            merged_row[j] = right_row[col.right_index]

    return merged_row


def _row_sort_key(keys: list[ColumnIndex]) -> Callable[[MicrodataRow], Any]:
    if len(keys) == 1:
        key = keys[0]
        return lambda row: row[key][MICRODATA_FLOAT_VALUE]
    else:
        return lambda row: [row[key][MICRODATA_FLOAT_VALUE] for key in keys]


def _merge_microdata(state: StitchState, left_rows: list[MicrodataRow], right_rows: list[MicrodataRow]) -> None:
    if len(left_rows) == 0 or len(right_rows) == 0:
        raise ValueError("Attempted a stitch with no rows.")

    context = state.context
    stitch_owner = context.stitch_owner
    rng = context.rng
    all_columns = context.all_columns

    if stitch_owner == StitchOwner.LEFT:
        num_rows = len(left_rows)
    elif stitch_owner == StitchOwner.RIGHT:
        num_rows = len(right_rows)
    else:
        num_rows = int(round((len(left_rows) + len(right_rows)) / 2.0))

    rng.shuffle(left_rows)
    left_rows = _align_length(rng, num_rows, left_rows)
    left_rows.sort(key=_row_sort_key(context.left_stitch_indexes))

    rng.shuffle(right_rows)
    right_rows = _align_length(rng, num_rows, right_rows)
    right_rows.sort(key=_row_sort_key(context.right_stitch_indexes))

    for i in range(num_rows):
        pick_shared_left = i % 2 == 0 if stitch_owner == StitchOwner.SHARED else stitch_owner == StitchOwner.LEFT
        context.result_rows.append(_merge_row(all_columns, pick_shared_left, left_rows[i], right_rows[i]))


THRESH_REL = 0.7


def _acceptable_distribution(left: int, right: int) -> bool:
    min_val = min(left, right)
    max_val = max(left, right)

    if min_val == 0:
        return False
    else:
        rel_diff = min_val / max_val
        return rel_diff >= THRESH_REL


def _can_split(state: StitchState) -> bool:
    context = state.context
    split_column = state.next_sort_column

    if context.stitch_is_integral[split_column]:
        interval = state.stitch_intervals[split_column]

        if context.stitch_max_values[split_column] == interval.max:
            return interval.size() >= 1.0
        else:
            return interval.size() > 1.0
    else:
        return True


def update_at(lst: list[Any], index: int, new_value: Any) -> list[Any]:
    if index < 0 or index >= len(lst):
        raise IndexError("Index out of range")

    modified_list = lst.copy()
    modified_list[index] = new_value
    return modified_list


def _stitch_rec(state: StitchState, left_rows: list[MicrodataRow], right_rows: list[MicrodataRow]) -> None:
    if state.remaining_sort_attempts == 0 or len(left_rows) == 1 or len(right_rows) == 1:
        _merge_microdata(state, left_rows, right_rows)
    elif _can_split(state):
        context = state.context
        current_sort_column = state.next_sort_column

        left_stitch_index = context.left_stitch_indexes[current_sort_column]
        right_stitch_index = context.right_stitch_indexes[current_sort_column]

        if state.currently_sorted_by != current_sort_column:
            left_rows.sort(key=_row_sort_key([left_stitch_index]))
            right_rows.sort(key=_row_sort_key([right_stitch_index]))

        interval = state.stitch_intervals[current_sort_column]
        interval_middle = interval.middle()

        left_split_point = max(0, _binary_search(left_rows, left_stitch_index, interval_middle, 0, len(left_rows)))
        right_split_point = max(0, _binary_search(right_rows, right_stitch_index, interval_middle, 0, len(right_rows)))

        left_lower = left_rows[:left_split_point]
        right_lower = right_rows[:right_split_point]

        left_upper = left_rows[left_split_point:]
        right_upper = right_rows[right_split_point:]

        if _acceptable_distribution(len(left_lower), len(right_lower)) and _acceptable_distribution(
            len(left_upper), len(right_upper)
        ):
            # Visit lower half.
            _stitch_rec(
                state.replace(
                    depth=state.depth + 1,
                    stitch_intervals=update_at(state.stitch_intervals, current_sort_column, interval.lower_half()),
                    next_sort_column=(current_sort_column + 1) % context.num_stitch_columns,
                    currently_sorted_by=current_sort_column,
                    remaining_sort_attempts=context.num_stitch_columns,
                ),
                left_lower,
                right_lower,
            )

            # Visit upper half.
            _stitch_rec(
                state.replace(
                    depth=state.depth + 1,
                    stitch_intervals=update_at(state.stitch_intervals, current_sort_column, interval.upper_half()),
                    next_sort_column=(current_sort_column + 1) % context.num_stitch_columns,
                    currently_sorted_by=current_sort_column,
                    remaining_sort_attempts=context.num_stitch_columns,
                ),
                left_upper,
                right_upper,
            )
        else:
            next_stitch_intervals = state.stitch_intervals
            if len(left_lower) == 0 and len(right_lower) == 0:
                next_stitch_intervals = update_at(state.stitch_intervals, current_sort_column, interval.upper_half())
            elif len(left_upper) == 0 and len(right_upper) == 0:
                next_stitch_intervals = update_at(state.stitch_intervals, current_sort_column, interval.lower_half())

            # Try next column.
            _stitch_rec(
                state.replace(
                    next_sort_column=(current_sort_column + 1) % context.num_stitch_columns,
                    stitch_intervals=next_stitch_intervals,
                    currently_sorted_by=current_sort_column,
                    remaining_sort_attempts=state.remaining_sort_attempts - 1,
                ),
                left_rows,
                right_rows,
            )
    else:
        # Try next column.
        _stitch_rec(
            state.replace(
                next_sort_column=(state.next_sort_column + 1) % state.context.num_stitch_columns,
                remaining_sort_attempts=state.remaining_sort_attempts - 1,
            ),
            left_rows,
            right_rows,
        )


def _do_stitch(
    forest: Forest,
    metadata: StitchingMetadata,
    left: tuple[list[MicrodataRow], Combination],
    right: tuple[list[MicrodataRow], Combination],
    derived_cluster: DerivedCluster,
) -> tuple[list[MicrodataRow], Combination]:
    (left_rows, left_combination) = left
    (right_rows, right_combination) = right
    (stitch_owner, stitch_columns, derived_columns) = derived_cluster

    if len(left_combination) == 0 or len(stitch_columns) == 0 or len(derived_columns) == 0:
        raise ValueError("Invalid clusters in stitch operation.")

    # Pick lowest entropy column first.
    stitch_columns = sorted(stitch_columns, key=lambda col: (metadata.entropy_1dim[col], col))
    all_columns = _locate_columns(left_combination, right_combination)

    if len(left_rows) == 0 and len(right_rows) == 0:
        return [], tuple(c.column_id for c in all_columns)

    if len(right_rows) == 0:
        raise ValueError(f"Empty sequence in cluster {right_combination}.")

    result_rows: list[MicrodataRow] = []

    root_stitch_intervals = [forest.snapped_intervals[col] for col in stitch_columns]

    stitch_state = StitchState(
        depth=0,
        stitch_intervals=root_stitch_intervals,
        next_sort_column=0,
        currently_sorted_by=-1,
        remaining_sort_attempts=len(stitch_columns),
        context=StitchContext(
            rng=forest.unsafe_rng,
            stitch_owner=stitch_owner,
            all_columns=all_columns,
            entropy_1dim=metadata.entropy_1dim[stitch_columns],
            stitch_max_values=[r.max for r in root_stitch_intervals],
            stitch_is_integral=[metadata.dimension_is_integral[col] for col in stitch_columns],
            left_stitch_indexes=_find_indexes(stitch_columns, left_combination),
            right_stitch_indexes=_find_indexes(stitch_columns, right_combination),
            result_rows=result_rows,
        ),
    )

    _stitch_rec(stitch_state, left_rows, right_rows)

    return result_rows, tuple(c.column_id for c in all_columns)


def _do_patch(
    rng: Random, left: tuple[list[MicrodataRow], Combination], right: tuple[list[MicrodataRow], Combination]
) -> tuple[list[MicrodataRow], Combination]:
    (left_rows, left_combination) = left
    (right_rows, right_combination) = right

    DOESNT_MATTER = True

    all_columns = _locate_columns(left_combination, right_combination)
    num_rows = len(left_rows)

    left_rows = left_rows[:]
    right_rows = right_rows[:]
    rng.shuffle(right_rows)
    right_rows = _align_length(rng, num_rows, right_rows)

    all_rows = [
        _merge_row(all_columns, DOESNT_MATTER, left_row, right_row)
        for left_row, right_row in zip(left_rows, right_rows)
    ]

    all_columns_ids = tuple(c.column_id for c in all_columns)
    return all_rows, all_columns_ids


def _stitch(
    materialize_tree: TreeMaterializer,
    forest: Forest,
    metadata: StitchingMetadata,
    left: tuple[list[MicrodataRow], Combination],
    derived_cluster: DerivedCluster,
) -> tuple[list[MicrodataRow], Combination]:
    (_, stitch_columns, derived_columns) = derived_cluster

    right = materialize_tree(forest, stitch_columns + derived_columns)

    if len(stitch_columns) == 0:
        return _do_patch(forest.unsafe_rng, left, right)
    else:
        return _do_stitch(forest, metadata, left, right, derived_cluster)


def build_table(
    materialize_tree: TreeMaterializer, forest: Forest, metadata: StitchingMetadata, clusters: Clusters
) -> tuple[list[Row], Combination]:
    acc = materialize_tree(forest, clusters.initial_cluster)

    for derived_cluster in clusters.derived_clusters:
        acc = _stitch(materialize_tree, forest, metadata, acc, derived_cluster)

    rows, columns = acc
    return [microdata_row_to_row(row) for row in rows], columns



from abc import ABC, abstractmethod

import numpy as np
import numpy.typing as npt

from .anonymizer import *
from .common import *

# If PID values are guaranteed to be unique, we don't need to track distinct values when checking
# whether entities are low-count or apply flattening when counting rows, which greatly reduces memory usage.
# In the generic case, tracking entities requires less memory than tracking row contributions,
# so the two counters are separated into different types.


class IEntityCounter(ABC):
    @abstractmethod
    def add(self, pids: Hashes) -> None:
        pass

    @abstractmethod
    def is_low_count(self, salt: bytes, params: SuppressionParams) -> bool:
        pass


class IRowCounter(ABC):
    @abstractmethod
    def add(self, pids: Hashes) -> None:
        pass

    @abstractmethod
    def noisy_count(self, context: AnonymizationContext) -> int:
        pass


class GenericPidEntityCounter(IEntityCounter):
    def __init__(self, dimensions: int, max_low_count: int) -> None:
        self.pid_sets: list[npt.NDArray[Hash]] = [np.zeros(max_low_count, Hash) for _ in range(dimensions)]
        self.pid_counts: list[int] = [0 for _ in range(dimensions)]
        self.max_low_count = max_low_count

    def add(self, pids: Hashes) -> None:
        for dimension, pid in enumerate(pids):
            pid_set = self.pid_sets[dimension]
            count = self.pid_counts[dimension]
            if count < self.max_low_count and pid != 0 and pid not in pid_set:
                pid_set[count] = pid
                self.pid_counts[dimension] = count + 1

    def is_low_count(self, salt: bytes, params: SuppressionParams) -> bool:
        pid_trackers = [
            (count, seed_from_pid_set(pids))
            for count, pids in zip(self.pid_counts, self.pid_sets)
            if count < self.max_low_count
        ]
        if pid_trackers == []:
            return False
        return is_low_count(salt, params, pid_trackers)


class GenericPidRowCounter(IRowCounter):
    def __init__(self, dimensions: int) -> None:
        self.contributions_list = [PidContributions() for _ in range(dimensions)]

    def add(self, pids: Hashes) -> None:
        for i, pid in enumerate(pids):
            contributions = self.contributions_list[i]
            if pid != 0:
                contributions.value_counts[pid] += 1
            else:
                # Missing PID value, add to the unaccounted rows count, so we can flatten them separately.
                contributions.unaccounted_for += 1

    def noisy_count(self, context: AnonymizationContext) -> int:
        result = count_multiple_contributions(context, self.contributions_list)
        return result.anonymized_count if result is not None else 0


class UniquePidCounter(IEntityCounter, IRowCounter):
    def __init__(self) -> None:
        self.real_count = 0
        self.seed = Hash(0)

    def add(self, pids: Hashes) -> None:
        assert len(pids) == 1

        if pids[0] != 0:
            self.real_count += 1
            self.seed ^= pids[0]

    def is_low_count(self, salt: bytes, params: SuppressionParams) -> bool:
        return is_low_count(salt, params, [(self.real_count, self.seed)])

    def noisy_count(self, context: AnonymizationContext) -> int:
        return count_single_contributions(context, self.real_count, self.seed)


class CountersFactory(ABC):
    @abstractmethod
    def create_entity_counter(self) -> IEntityCounter:
        pass

    @abstractmethod
    def create_row_counter(self) -> IRowCounter:
        pass


class UniquePidCountersFactory(CountersFactory):
    def create_row_counter(self) -> IRowCounter:
        return UniquePidCounter()

    def create_entity_counter(self) -> IEntityCounter:
        return UniquePidCounter()


class GenericPidCountersFactory(CountersFactory):
    def __init__(self, dimensions: int, max_low_count: int) -> None:
        self.dimensions = dimensions
        self.max_low_count = max_low_count

    def create_row_counter(self) -> IRowCounter:
        return GenericPidRowCounter(self.dimensions)

    def create_entity_counter(self) -> IEntityCounter:
        return GenericPidEntityCounter(self.dimensions, self.max_low_count)



import os
import secrets
import sys
from dataclasses import replace
from typing import Optional

import pandas as pd
from appdirs import user_config_dir

from .bucket import harvest
from .clustering.common import MicrodataRow
from .clustering.stitching import StitchingMetadata, build_table
from .clustering.strategy import ClusteringStrategy, DefaultClustering, MlClustering
from .common import *
from .counters import (
    CountersFactory,
    GenericPidCountersFactory,
    UniquePidCountersFactory,
)
from .forest import Forest
from .microdata import apply_convertors, generate_microdata, get_convertor


def _get_default_salt() -> bytes:
    config_dir = user_config_dir("SynDiffix", "OpenDiffix")
    salt_file_path = os.path.join(config_dir, "salt.bin")

    try:
        if not os.path.isfile(salt_file_path):
            salt = secrets.randbits(64).to_bytes(8, "little")

            os.makedirs(config_dir, exist_ok=True)
            with open(salt_file_path, "wb") as file:
                file.write(salt)

        with open(salt_file_path, "rb") as file:
            return file.read()
    except Exception as e:
        print(
            "Error: could not create or retrieve the default salt value!\n"
            + f"Make sure you have the required rights to access the file `{salt_file_path}` "
            + "or provide a custom salt value.",
            file=sys.stderr,
        )
        raise e


class Synthesizer(object):
    @staticmethod
    def _is_integral(col_type: ColumnType) -> bool:
        if col_type == ColumnType.REAL or col_type == ColumnType.TIMESTAMP:
            return False
        else:
            return True

    def __init__(
        self,
        raw_data: pd.DataFrame,
        pids: Optional[pd.DataFrame] = None,
        anonymization_params: AnonymizationParams = AnonymizationParams(),
        bucketization_params: BucketizationParams = BucketizationParams(),
        target_column: Optional[ColumnId | str] = None,
        clustering: Optional[ClusteringStrategy] = None,
    ) -> None:
        if target_column is not None:
            if clustering:
                raise ValueError("Cannot specify both target_column and clustering parameters.")
            clustering = MlClustering(target_column=target_column)
        elif not clustering:
            clustering = DefaultClustering()

        if anonymization_params.salt == b"":
            anonymization_params = replace(anonymization_params, salt=_get_default_salt())

        if pids is None:
            pids = pd.DataFrame({"RowIndex": range(1, len(raw_data) + 1)})
            counters_factory: CountersFactory = UniquePidCountersFactory()
        else:
            low_count_params = anonymization_params.low_count_params
            # Stop counting entities over 4 standard deviations more than the mean of the range threshold.
            # `low_mean_gap` is the number of standard deviations between `low_threshold` and desired mean.
            max_low_count = bucketization_params.range_low_threshold + int(
                (low_count_params.low_mean_gap + 4.0) * low_count_params.layer_sd
            )
            counters_factory = GenericPidCountersFactory(len(pids.columns), max_low_count)

        self.raw_dtypes = raw_data.dtypes

        self.column_convertors = [get_convertor(raw_data, column) for column in raw_data.columns]
        self.column_is_integral = [self._is_integral(convertor.column_type()) for convertor in self.column_convertors]

        self.forest = Forest(
            anonymization_params,
            bucketization_params,
            counters_factory,
            pids,
            apply_convertors(self.column_convertors, raw_data),
        )

        self.clusters, self.entropy_1dim = clustering.build_clusters(self.forest)
        for col_id, converter in enumerate(self.column_convertors):
            converter.analyze_tree(self.forest.get_tree((ColumnId(col_id),)))

    def sample(self) -> pd.DataFrame:
        def materialize_tree(forest: Forest, columns: list[ColumnId]) -> tuple[list[MicrodataRow], Combination]:
            combination = tuple(sorted(columns))
            tree = forest.get_tree(combination)
            buckets = harvest(tree, self.forest.derive_unsafe_rng())
            return (
                generate_microdata(
                    buckets,
                    get_items_combination_list(combination, self.column_convertors),
                    get_items_combination_list(combination, forest.null_mappings),
                    forest.derive_unsafe_rng(),
                ),
                combination,
            )

        rows, root_combination = build_table(
            materialize_tree,
            self.forest,
            StitchingMetadata(self.column_is_integral, self.entropy_1dim),
            self.clusters,
        )

        syn_data = pd.DataFrame(rows, columns=get_items_combination(root_combination, self.forest.columns))
        # Convert the new columns to their original type. We need to account that some columns might be missing.
        syn_data = syn_data.astype({column: self.raw_dtypes[column] for column in syn_data.columns}, copy=False)

        return syn_data

    @property
    def salt(self) -> bytes:
        return self.forest.anonymization_params.salt



import math
from dataclasses import dataclass

from ..common import *
from .common import *


@dataclass
class MutableCluster:
    columns: set[ColumnId]
    total_entropy: float


DERIVED_COLS_MIN = 1
DERIVED_COLS_RESERVED = 0.5
DERIVED_COLS_RATIO = 0.7


def _col_weight(entropy: float) -> float:
    return 1.0 + math.sqrt(max(entropy, 1.0))


def _floor_by(value: float, bin_size: float) -> float:
    return math.floor(value / bin_size) * bin_size


def _build_clusters(
    context: ClusteringContext,
    max_weight: float,
    merge_thresh: float,
    col_weights: list[float],
    permutation: list[ColumnId],
) -> Clusters:
    dependency_matrix = context.dependency_matrix

    clusters: list[MutableCluster] = []
    if context.main_column is not None:
        clusters = [MutableCluster(columns={context.main_column}, total_entropy=col_weights[context.main_column])]
        permutation = [col for col in permutation if col != context.main_column]

    # For each column in the permutation, we find the "best" cluster that has available space.
    # We judge how suitable a cluster is by average dependence score.
    # If no available cluster is found, we start a new one.
    # After all clusters are built, we fill remaining space of non-initial clusters with stitch columns.

    for col in permutation:
        best_cluster = None
        best_avg_quality = -1.0

        for i, cluster in enumerate(clusters):
            capacity = max_weight if i == 0 else DERIVED_COLS_RATIO * max_weight
            average_quality = sum(dependency_matrix[col, c] for c in cluster.columns) / len(cluster.columns)

            # Skip if below threshold or above weight limit.
            if average_quality < merge_thresh or (
                len(cluster.columns) > DERIVED_COLS_MIN and cluster.total_entropy + col_weights[col] > capacity
            ):
                continue

            if average_quality > best_avg_quality:
                # Found new best cluster.
                best_cluster = cluster
                best_avg_quality = average_quality

        if best_cluster is not None:
            best_cluster.columns.add(col)
            best_cluster.total_entropy += col_weights[col]
        else:
            clusters.append(MutableCluster(columns={col}, total_entropy=col_weights[col]))

    derived_clusters: list[DerivedCluster] = []
    available_columns = clusters[0].columns.copy()

    for i in range(1, len(clusters)):
        cluster = clusters[i]
        total_weight = max(cluster.total_entropy, DERIVED_COLS_RESERVED * max_weight)

        stitch_columns: set[ColumnId] = set()
        derived_columns = list(cluster.columns)

        best_stitch_columns = [
            (
                c_left,
                sum(dependency_matrix[c_left, c_right] for c_right in derived_columns) / len(derived_columns),
                max(dependency_matrix[c_left, c_right] for c_right in derived_columns),
            )
            for c_left in available_columns
        ]

        best_stitch_columns.sort(
            key=lambda x: (_floor_by(x[1], 0.05), _floor_by(x[2], 0.01), context.total_dependence_per_column[x[0]]),
            reverse=True,
        )

        best_stitch_col = context.main_column if context.main_column is not None else best_stitch_columns[0][0]
        stitch_columns.add(best_stitch_col)
        total_weight += col_weights[best_stitch_col]

        for c_left, _dep_avg, dep_max in best_stitch_columns:
            if c_left != best_stitch_col and dep_max >= merge_thresh:
                weight = col_weights[c_left]
                if total_weight + weight <= max_weight:
                    stitch_columns.add(c_left)
                    total_weight += weight

        available_columns.update(derived_columns)
        derived_clusters.append((StitchOwner.SHARED, list(stitch_columns), derived_columns))

    initial_cluster = list(clusters[0].columns)

    return Clusters(initial_cluster=initial_cluster, derived_clusters=derived_clusters)


def _clustering_quality(context: ClusteringContext, clusters: Clusters) -> float:
    dependency_matrix = context.dependency_matrix

    unsatisfied_dependence = context.total_dependence

    def visit_pairs(columns: list[ColumnId]) -> None:
        nonlocal unsatisfied_dependence

        for i in range(1, len(columns)):
            col_a = columns[i]

            for j in range(i):
                col_b = columns[j]
                unsatisfied_dependence -= dependency_matrix[col_a, col_b]
                unsatisfied_dependence -= dependency_matrix[col_b, col_a]

    visit_pairs(clusters.initial_cluster)

    for _, stitch_columns, derived_columns in clusters.derived_clusters:
        visit_pairs(stitch_columns + derived_columns)

    return unsatisfied_dependence / (2.0 * context.num_columns)


def _simplify_clusters(clusters: Clusters) -> Clusters:
    if len(clusters.derived_clusters) == 0:
        return clusters

    # If we have clusters in shape initial=[a], derived=[[a],[b,c]]..., we merge to initial=[a,b,c].
    derived_cluster = clusters.derived_clusters[0]
    if set(clusters.initial_cluster) == set(derived_cluster[1]):
        return Clusters(
            initial_cluster=derived_cluster[1] + derived_cluster[2], derived_clusters=clusters.derived_clusters[1:]
        )

    return clusters


def _do_solve(context: ClusteringContext, max_weight: float, merge_thresh: float, alpha: float) -> Clusters:
    num_cols = context.num_columns
    rng = context.rng

    # Constants
    initial_solution = [ColumnId(i) for i in range(num_cols)]

    initial_temperature = 5.0
    min_temperature = 3.5e-3

    def next_temperature(current_temp: float) -> float:
        return current_temp / (1.0 + alpha * current_temp)

    def mutate(solution: list[ColumnId]) -> list[ColumnId]:
        copy = solution.copy()
        i = rng.randint(0, num_cols - 1)
        j = rng.randint(0, num_cols - 1)
        while i == j:
            j = rng.randint(0, num_cols - 1)

        copy[i], copy[j] = solution[j], solution[i]
        return copy

    col_weights = list(_col_weight(col) for col in context.entropy_1dim)

    def evaluate(solution: list[ColumnId]) -> float:
        clusters = _build_clusters(context, max_weight, merge_thresh, col_weights, solution)
        return _clustering_quality(context, clusters)

    # Solver state
    current_solution = initial_solution
    current_energy = evaluate(initial_solution)
    best_solution = initial_solution
    best_energy = current_energy
    temperature = initial_temperature

    # Simulated annealing loop
    while best_energy > 0 and temperature > min_temperature:
        new_solution = mutate(current_solution)
        new_energy = evaluate(new_solution)
        energy_delta = new_energy - current_energy

        if energy_delta <= 0.0 or math.exp(-energy_delta / temperature) > rng.random():
            current_solution = new_solution
            current_energy = new_energy

        if current_energy < best_energy:
            best_solution = current_solution
            best_energy = current_energy

        temperature = next_temperature(temperature)

    return _simplify_clusters(_build_clusters(context, max_weight, merge_thresh, col_weights, best_solution))


def solve(context: ClusteringContext, max_weight: float, merge_thresh: float, alpha: float) -> Clusters:
    assert max_weight > 1.0
    num_cols = context.num_columns

    if num_cols <= 4:
        # Build a cluster that includes everything.
        return Clusters(initial_cluster=[ColumnId(i) for i in range(num_cols)], derived_clusters=[])

    # TODO: Do an exact search up to a number of columns.
    return _do_solve(context, max_weight, merge_thresh, alpha)


def solve_with_features(
    main_column: ColumnId,
    main_features: list[ColumnId],
    max_weight: float,
    entropy_1dim: Entropy1Dim,
    drop_non_features: bool,
) -> Clusters:
    num_columns = len(entropy_1dim)
    main_column_weight = _col_weight(entropy_1dim[main_column])

    clusters: list[MutableCluster] = []

    def new_cluster() -> MutableCluster:
        cluster = MutableCluster(columns=set(), total_entropy=main_column_weight)
        clusters.append(cluster)
        return cluster

    curr = new_cluster()
    curr.columns.add(main_column)  # Only for the first cluster, in others main is a stitch column.

    for feature in main_features:
        weight = _col_weight(entropy_1dim[feature])

        if len(curr.columns) > 1 and curr.total_entropy + weight > max_weight:
            curr = new_cluster()

        curr.columns.add(feature)
        curr.total_entropy += weight

    initial_cluster = list(clusters[0].columns)
    derived_clusters = [(StitchOwner.SHARED, [main_column], list(cluster.columns)) for cluster in clusters[1:]]

    ml_columns = [main_column] + main_features
    non_feature_columns: list[DerivedCluster] = (
        []
        if drop_non_features
        else [(StitchOwner.LEFT, [main_column], [ColumnId(c)]) for c in range(num_columns) if c not in ml_columns]
    )

    return Clusters(initial_cluster=initial_cluster, derived_clusters=derived_clusters + non_feature_columns)



from dataclasses import dataclass
from itertools import chain
from math import floor
from random import Random

from .common import *
from .interval import Interval, Intervals
from .tree import Branch, Leaf, Node


@dataclass
class Bucket:
    intervals: Intervals
    count: int


Buckets = list[Bucket]

BucketsCache = dict[Node, Buckets]


def _adjust_counts(buckets: Buckets, current_count: int, target_count: int) -> None:
    ratio = target_count / current_count

    acc_error = 0.0
    for bucket in buckets:
        adjusted_count = bucket.count * ratio
        acc_error += adjusted_count - floor(adjusted_count)

        # Add the previously accumulated errors to the current count, if larger than 1 unit.
        if acc_error > 1.0:
            adjusted_count += 1.0
            acc_error -= 1.0

        bucket.count = int(adjusted_count)


def _get_subbuckets(node: Node, harvested_nodes: BucketsCache, unsafe_rng: Random) -> list[Buckets]:
    subnodes_combinations = generate_combinations(node.dimensions() - 1, node.dimensions())
    subnodes_buckets: list[Buckets] = []

    for subnode, combination in zip(node.subnodes, subnodes_combinations):
        actual_subnode_intervals = get_items_combination(combination, node.actual_intervals)
        if all(interval.is_singularity() for interval in actual_subnode_intervals):
            subnode_buckets = [Bucket(actual_subnode_intervals, node.noisy_count())]
        elif subnode is None:
            subnode_buckets = []
        else:
            subnode_buckets = _harvest_node(subnode, harvested_nodes, unsafe_rng)

        subnodes_buckets.append(subnode_buckets)

    return subnodes_buckets


def _get_smallest_intervals(subbuckets: list[Buckets]) -> Intervals:
    dimensions = len(subbuckets)
    cumulative_intervals: list[list[Interval | None]] = [[None] * dimensions for _ in range(dimensions)]

    for combination_index, combination in enumerate(generate_combinations(dimensions - 1, dimensions)):
        for bucket in subbuckets[combination_index]:
            for interval_index, bucket_interval in enumerate(bucket.intervals):
                dimension_index = combination[interval_index]

                match cumulative_intervals[dimension_index][combination_index]:
                    case None:
                        # This gets mutated later, so we need to create a new Interval instance.
                        cumulative_intervals[dimension_index][combination_index] = bucket_interval.copy()
                    case cumulative_interval:
                        cumulative_interval = cast(Interval, cumulative_interval)  # Needed to silence the type checker.
                        cumulative_interval.min = min(cumulative_interval.min, bucket_interval.min)
                        cumulative_interval.max = max(cumulative_interval.max, bucket_interval.max)

    return tuple(
        min(filter(None, intervals), key=lambda interval: interval.size()) for intervals in cumulative_intervals
    )


def _get_per_dimension_interval_lists(smallest_intervals: Intervals, subbuckets: list[Buckets]) -> list[list[Interval]]:
    dimensions = len(subbuckets)
    per_dimension_interval_lists: list[list[Interval]] = [[] for _ in range(dimensions)]

    for subnode_buckets, subnode_combination in zip(subbuckets, generate_combinations(dimensions - 1, dimensions)):
        for interval_index, dimension_index in enumerate(subnode_combination):
            smallest_interval = smallest_intervals[dimension_index]
            for bucket in subnode_buckets:
                bucket_interval = bucket.intervals[interval_index]
                if smallest_interval.contains_interval(bucket_interval):
                    per_dimension_interval_lists[dimension_index].extend([bucket_interval] * bucket.count)

    return per_dimension_interval_lists


#  Removes the low-count half, if one exists, from the specified interval.
def _compact_node_interval(leaf: Leaf, dimension: int, interval: Interval) -> Interval:
    value_index = leaf.context.combination[dimension]

    # IEntityCounter is mutable, so we need to create a new object for each half.
    per_half_entity_counters = (
        leaf.context.counters_factory.create_entity_counter(),
        leaf.context.counters_factory.create_entity_counter(),
    )

    for row in leaf.rows:
        half_index = interval.half_index(leaf.context.data[row][value_index])
        per_half_entity_counters[half_index].add(leaf.context.pid_data[row])

    anon_params = leaf.context.anonymization_context.anonymization_params
    per_half_low_count = tuple(
        counter.is_low_count(anon_params.salt, anon_params.low_count_params) for counter in per_half_entity_counters
    )
    match per_half_low_count:
        case (False, True):
            return interval.half(0)
        case (True, False):
            return interval.half(1)
        case _:
            return interval


def _compact_node_intervals(node: Node) -> Intervals:
    if isinstance(node, Leaf) and node.is_stub:
        return tuple(
            _compact_node_interval(node, index, interval) for index, interval in enumerate(node.snapped_intervals)
        )
    else:
        return node.snapped_intervals


def _compact_smallest_intervals(node: Node, smallest_intervals: Intervals) -> None:
    compacted_node_intervals = _compact_node_intervals(node)
    for smallest_interval, compacted_node_interval in zip(smallest_intervals, compacted_node_intervals):
        if compacted_node_interval.overlaps(smallest_interval):
            # Drop the sections of the smallest interval which are outside the compacted node interval.
            smallest_interval.min = max(smallest_interval.min, compacted_node_interval.min)
            smallest_interval.max = min(smallest_interval.max, compacted_node_interval.max)
        else:
            # If the intervals have no overlap, expand the smallest interval to include both.
            smallest_interval.min = min(smallest_interval.min, compacted_node_interval.min)
            smallest_interval.max = max(smallest_interval.max, compacted_node_interval.max)


def _get_per_subnode_intervals_lists(smallest_intervals: Intervals, subbuckets: list[Buckets]) -> list[list[Intervals]]:
    assert len(smallest_intervals) == len(subbuckets)
    dimensions = len(smallest_intervals)
    per_subnode_intervals_lists: list[list[Intervals]] = [[] for _ in range(dimensions)]

    for combination_index, combination in enumerate(generate_combinations(dimensions - 1, dimensions)):
        # For every subnode, gather only the subintervals that are contained in the corresponding smallest intervals.
        subsmallest_intervals = get_items_combination(combination, smallest_intervals)
        for subbucket in subbuckets[combination_index]:
            if all(subsmallest_intervals[i].contains_interval(subbucket.intervals[i]) for i in range(dimensions - 1)):
                per_subnode_intervals_lists[combination_index].extend([subbucket.intervals] * subbucket.count)

    return per_subnode_intervals_lists


def _match_subintervals(
    count: int,
    per_dimension_interval_lists: list[list[Interval]],
    per_subnode_intervals_lists: list[list[Intervals]],
    unsafe_rng: Random,
) -> Iterable[Intervals]:
    assert len(per_dimension_interval_lists) == len(per_subnode_intervals_lists)
    dimensions = len(per_dimension_interval_lists)

    for match_counter in range(count):
        # We cycle through all the dimensions and subnodes equally when selecting the sub-intervals to be matched.
        subnode_index = match_counter % dimensions
        dimension_index = dimensions - subnode_index - 1

        subnode_intervals_list = per_subnode_intervals_lists[subnode_index]
        dimension_interval_list = per_dimension_interval_lists[dimension_index]

        # Create a random match between the selected sub-intervals in order to avoid biases in the output.
        subnode_intervals = subnode_intervals_list[unsafe_rng.randint(0, len(subnode_intervals_list) - 1)]
        dimension_interval = dimension_interval_list[unsafe_rng.randint(0, len(dimension_interval_list) - 1)]

        yield (*subnode_intervals[:dimension_index], dimension_interval, *subnode_intervals[dimension_index:])


def _refine_buckets(node: Node, harvested_nodes: BucketsCache, count: int, unsafe_rng: Random) -> Buckets:
    subbuckets = _get_subbuckets(node, harvested_nodes, unsafe_rng)

    if not all(subbuckets):
        return [Bucket(tuple(node.bucket_intervals()), count)]

    smallest_intervals = _get_smallest_intervals(subbuckets)
    _compact_smallest_intervals(node, smallest_intervals)
    per_dimension_interval_lists = _get_per_dimension_interval_lists(smallest_intervals, subbuckets)
    per_subnode_intervals_lists = _get_per_subnode_intervals_lists(smallest_intervals, subbuckets)

    if not all(per_dimension_interval_lists) or not all(per_subnode_intervals_lists):
        # Can't match if there are no intervals for some subnodes or no interval for some dimensions.
        return [Bucket(tuple(node.bucket_intervals()), count)]

    matched_intervals = _match_subintervals(
        count, per_dimension_interval_lists, per_subnode_intervals_lists, unsafe_rng
    )
    return [Bucket(interval, 1) for interval in matched_intervals]


def _harvest_leaf(leaf: Leaf, harvested_nodes: BucketsCache, unsafe_rng: Random) -> Buckets:
    low_threshold = leaf.context.anonymization_context.anonymization_params.low_count_params.low_threshold
    if leaf.is_over_threshold(low_threshold):
        if leaf.is_singularity() or leaf.dimensions() == 1:
            return [Bucket(tuple(leaf.bucket_intervals()), leaf.noisy_count())]
        else:
            return _refine_buckets(leaf, harvested_nodes, leaf.noisy_count(), unsafe_rng)
    else:
        return []


def _harvest_branch(branch: Branch, harvested_nodes: BucketsCache, unsafe_rng: Random) -> Buckets:
    children_buckets = list(
        chain.from_iterable(
            _harvest_node(child_node, harvested_nodes, unsafe_rng) for child_node in branch.children.values()
        )
    )

    children_count = sum(bucket.count for bucket in children_buckets)
    parent_count = branch.noisy_count()

    if 2 * children_count < parent_count:
        if branch.dimensions() == 1:
            children_buckets = [Bucket(tuple(branch.bucket_intervals()), parent_count)]
        else:
            children_buckets += _refine_buckets(branch, harvested_nodes, parent_count - children_count, unsafe_rng)
    else:
        _adjust_counts(children_buckets, children_count, parent_count)

    return children_buckets


def _harvest_node(node: Node, harvested_nodes: BucketsCache, unsafe_rng: Random) -> Buckets:
    if harvested_nodes.get(node) is None:
        harvester = _harvest_leaf if isinstance(node, Leaf) else _harvest_branch
        harvested_nodes[node] = harvester(node, harvested_nodes, unsafe_rng)  # type: ignore

    return harvested_nodes[node]


# The passed RNG isn't safe for anonymization, so it can only be applied to already anonymized data.
# Used for randomly matching subintervals during refining.
def harvest(node: Node, unsafe_rng: Random) -> Buckets:
    # Maps node.NodeData.Id to its already harvested buckets, so it's done only once.
    harvested_nodes: BucketsCache = dict()
    buckets = _harvest_node(node, harvested_nodes, unsafe_rng)
    return [bucket for bucket in buckets if bucket.count > 0]



import random
from dataclasses import dataclass
from enum import Enum, unique
from typing import Callable, NewType, Optional

import numpy.typing as npt

from ..common import *
from ..forest import Forest
from ..microdata import MICRODATA_SYN_VALUE, MicrodataRow

# Local index of a column in a microtable.
ColumnIndex = NewType("ColumnIndex", int)


@unique
class StitchOwner(Enum):
    LEFT = 0
    RIGHT = 1
    SHARED = 2

    def __repr__(self) -> str:
        return self.name


# Owner, Stitch columns, Derived columns
DerivedCluster = tuple[StitchOwner, list[ColumnId], list[ColumnId]]


@dataclass(frozen=True)
class Clusters:
    initial_cluster: list[ColumnId]
    derived_clusters: list[DerivedCluster]


TreeMaterializer = Callable[[Forest, list[ColumnId]], tuple[list[MicrodataRow], Combination]]


def microdata_row_to_row(microdata_row: MicrodataRow) -> Row:
    return tuple(value[MICRODATA_SYN_VALUE] for value in microdata_row)


Entropy1Dim = npt.NDArray[np.float64]


@dataclass(frozen=True)
class ClusteringContext:
    dependency_matrix: npt.NDArray[np.float64]
    entropy_1dim: Entropy1Dim
    total_dependence: float
    total_dependence_per_column: list[float]
    anonymization_params: AnonymizationParams
    bucketization_params: BucketizationParams
    rng: random.Random
    main_column: Optional[ColumnId]

    @property
    def num_columns(self) -> int:
        return len(self.dependency_matrix)


@dataclass(frozen=True)
class StitchingMetadata:
    dimension_is_integral: list[bool]
    entropy_1dim: Entropy1Dim



from __future__ import annotations

import math
import random

import numpy as np
import numpy.typing as npt
from pandas import DataFrame
from pandas.api.types import is_float_dtype, is_integer_dtype, is_string_dtype

from .anonymizer import hash_pid, noisy_row_limit
from .common import *
from .interval import Interval, get_null_mapping, snap_interval
from .tree import *


def _dimension_interval(data: DataFrame, dimension: int) -> Interval:
    min = data.iloc[:, dimension].min()
    max = data.iloc[:, dimension].max()
    if math.isnan(min):
        assert math.isnan(max)
        min = max = 0.0
    return Interval(min, max)


class Forest:
    def __init__(
        self,
        anonymization_params: AnonymizationParams,
        bucketization_params: BucketizationParams,
        counters_factory: CountersFactory,
        pids: DataFrame,
        data: DataFrame,
    ) -> None:
        self.anonymization_params = anonymization_params
        self.bucketization_params = bucketization_params
        self.counters_factory = counters_factory
        self.orig_pids = pids
        self.orig_data = data
        self.unsafe_rng = random.Random(0)

        assert len(pids) == len(data)
        assert len(data) > 0
        assert len(pids.columns) >= 1
        for dtype in pids.dtypes:
            assert is_string_dtype(dtype) or is_integer_dtype(dtype)
        # We assume `microdata.apply_convertors` has been called on the data.
        for dtype in data.dtypes:
            assert is_float_dtype(dtype)

        self.columns = tuple(data.columns)
        self.dimensions = len(self.columns)

        actual_intervals = tuple(_dimension_interval(data, i) for i in range(self.dimensions))
        self.null_mappings = tuple(get_null_mapping(interval) for interval in actual_intervals)
        for interval, null_mapping in zip(actual_intervals, self.null_mappings):
            interval.expand(null_mapping)
        self.snapped_intervals = tuple(snap_interval(interval) for interval in actual_intervals)

        # Hash and store PID values.
        pids = cast(DataFrame, pids.map(hash_pid))  # type: ignore
        self.pid_data: npt.NDArray[np.uint64] = pids.to_numpy(Hash)

        # Arrange data in a numpy array, applying the null mappings to missing values.
        # `DataFrame.fillna` doesn't seem to accept the fill values by index, must build dict.
        null_mappings_dict = dict(zip(data.columns, self.null_mappings))
        self.data: npt.NDArray[np.float64] = data.fillna(null_mappings_dict, axis=0).to_numpy(np.float64)

        self._tree_cache: dict[Combination, Node] = {}

        # We need to flatten uni-dimensional trees, by pushing the root down as long as one of its halves
        # fails LCF, and update the corresponding dimension's interval, in order to anonymize its boundaries.
        snapped_intervals = list(self.snapped_intervals)
        for i in range(self.dimensions):
            combination = (ColumnId(i),)
            tree = self._build_tree(combination).push_down_1dim_root()
            snapped_intervals[i] = tree.snapped_intervals[0]
            self._tree_cache[combination] = tree  # Cache the flattened version of the tree.
        self.snapped_intervals = tuple(snapped_intervals)

    def _get_subnodes(self, upper_combination: Combination) -> Subnodes:
        dimensions = len(upper_combination)
        sub_combinations = generate_combinations(dimensions - 1, dimensions)
        return tuple(
            self.get_tree(get_items_combination(sub_combination, upper_combination))
            for sub_combination in sub_combinations
        )

    def _build_tree(self, combination: Combination) -> Node:
        subnodes = self._get_subnodes(combination)

        # Hash column names into the tree's base bucket seed.
        base_seed = hash_strings(map(str, get_items_combination(combination, self.columns)))
        anonymization_context = AnonymizationContext(base_seed, self.anonymization_params)

        row_limit = (
            noisy_row_limit(
                anonymization_context.anonymization_params.salt,
                anonymization_context.bucket_seed,
                len(self.data),
                self.bucketization_params.precision_limit_row_fraction,
            )
            if len(combination) == 1
            else 0  # Only 1dim trees are depth limited.
        )

        context = Context(
            combination,
            self.pid_data,
            self.data,
            anonymization_context,
            self.bucketization_params,
            row_limit,
            self.counters_factory,
        )

        root_intervals = get_items_combination(combination, self.snapped_intervals)
        tree: Node = Leaf(context, subnodes, root_intervals, RowId(0))
        for index in range(1, len(self.data)):
            tree = tree.add_row(0, RowId(index))
        return tree

    def get_tree(self, combination: Combination) -> Node:
        tree = self._tree_cache.get(combination)
        if tree is None:
            tree = self._build_tree(combination)
            self._tree_cache[combination] = tree
        return tree

    def derive_unsafe_rng(self) -> random.Random:
        return random.Random(self.unsafe_rng.random())



from dataclasses import replace

from ..common import *
from ..forest import Forest


def should_sample(forest: Forest, sample_size: int) -> bool:
    dimensions = forest.dimensions
    num_rows = len(forest.data)

    if sample_size >= num_rows:
        return False

    sampled_2dim_work = dimensions * dimensions * sample_size
    full_2dim_work = (num_rows * dimensions * 3) // 2

    total_work_with_sample = sampled_2dim_work + full_2dim_work
    total_work_without_sample = dimensions * dimensions * num_rows

    return total_work_without_sample > total_work_with_sample * 2


def sample_forest(forest: Forest, sample_size: int) -> Forest:
    sampling_anon_params = replace(
        forest.anonymization_params,
        low_count_params=SuppressionParams(low_threshold=2, layer_sd=0.5, low_mean_gap=1.0),
        layer_noise_sd=0.0,
    )

    # New RNG prevents `forest.Random` from being affected by sample size.
    rng = forest.derive_unsafe_rng()

    random_indices = rng.sample(range(len(forest.orig_data)), sample_size)

    sampled_pids = forest.orig_pids.iloc[random_indices]
    sampled_data = forest.orig_data.iloc[random_indices]

    return Forest(
        sampling_anon_params,
        forest.bucketization_params,
        forest.counters_factory,
        sampled_pids,
        sampled_data,
    )



from collections.abc import Iterable
from dataclasses import dataclass
from enum import Enum, unique
from itertools import combinations
from typing import NewType, Sequence, TypeVar, cast

import numpy as np
import numpy.typing as npt
import pandas as pd

Hash = np.uint64
Hashes = npt.NDArray[Hash]

Value = int | float | str | bool | pd.Timestamp | None
Row = tuple[Value, ...]


@unique
class ColumnType(Enum):
    BOOLEAN = 1
    INTEGER = 2
    REAL = 3
    STRING = 4
    TIMESTAMP = 5


@dataclass(frozen=True)
class Column:
    name: str
    type: ColumnType


Columns = tuple[Column, ...]


@dataclass(frozen=True)
class FlatteningInterval:
    lower: int = 2
    upper: int = 5


@dataclass(frozen=True)
class SuppressionParams:
    low_threshold: int = 3
    layer_sd: float = 1.0
    low_mean_gap: float = 2.0


@dataclass(frozen=True)
class BucketizationParams:
    singularity_low_threshold: int = 5
    range_low_threshold: int = 15
    precision_limit_row_fraction: int = 10000
    precision_limit_depth_threshold: int = 15


@dataclass(frozen=True)
class AnonymizationParams:
    # Each noise layer seed is salted before being hashed with a cryptographically-strong algorithm.
    # The salt value needs to have at least 64 bits of entropy (equal or higher than that of the seed).
    # If the provided salt is empty, a per-system salt will be generated and used.
    salt: bytes = b""
    low_count_params: SuppressionParams = SuppressionParams()
    outlier_count: FlatteningInterval = FlatteningInterval()
    top_count: FlatteningInterval = FlatteningInterval()
    layer_noise_sd: float = 1.0


@dataclass(frozen=True)
class AnonymizationContext:
    bucket_seed: Hash
    anonymization_params: AnonymizationParams


# Global index of a column.
ColumnId = NewType("ColumnId", int)

Combination = tuple[ColumnId, ...]


def generate_combinations(k: int, n: int) -> Iterable[Combination]:
    return cast(Iterable[Combination], combinations(range(n), k) if k > 0 else [])


T = TypeVar("T")


def get_items_combination(combination: Combination, items: Sequence[T]) -> tuple[T, ...]:
    return tuple(items[index] for index in combination)


def get_items_combination_list(combination: Combination, items: Sequence[T]) -> list[T]:
    return [items[index] for index in combination]



from typing import Tuple

import numpy as np
import pandas as pd

from .clustering.common import MicrodataRow, StitchOwner
from .clustering.measures import measure_entropy
from .clustering.stitching import StitchingMetadata, _do_stitch
from .clustering.strategy import NoClustering
from .common import ColumnId, Combination
from .synthesizer import Synthesizer


def _make_synthesizers(
    df_left: pd.DataFrame,
    df_right: pd.DataFrame,
    col_names_stitch: list[str],
) -> Tuple[Synthesizer, Synthesizer, Synthesizer]:
    syn_left = Synthesizer(df_left, clustering=NoClustering())
    syn_right = Synthesizer(df_right, clustering=NoClustering())
    df_left_short = df_left.iloc[[0]]
    df_right_short = df_right.iloc[[0]].copy()
    for col_name in col_names_stitch:
        df_right_short.rename(columns={col_name: f"__fake__{col_name}"}, inplace=True)
    syn_fake = Synthesizer(pd.concat([df_left_short, df_right_short], axis=1), clustering=NoClustering())
    syn_fake.forest.snapped_intervals = syn_left.forest.snapped_intervals + syn_right.forest.snapped_intervals
    return syn_fake, syn_left, syn_right


def _make_microdata(
    df: pd.DataFrame,
    syn: Synthesizer,
    columns: list[ColumnId],
) -> tuple[list[MicrodataRow], Combination]:
    col_names = list(df.columns)
    microdata: list[MicrodataRow] = []

    for i in range(len(df)):
        row = []
        for col_name in col_names:
            row.append((df[col_name].iloc[i], float(syn.forest.orig_data[col_name].iloc[i])))
        microdata.append(row)
    return (microdata, tuple(columns))


def stitch(df_left: pd.DataFrame, df_right: pd.DataFrame, shared: bool = True) -> pd.DataFrame:
    # Make the needed column names with the original names (not later rename)
    col_names_left = list(df_left.columns)
    col_names_right = list(df_right.columns)
    col_names_stitch = list(set(col_names_left) & set(col_names_right))
    col_names_right_minus_stitch = [col for col in col_names_right if col not in col_names_stitch]

    syn_fake, syn_left, syn_right = _make_synthesizers(df_left, df_right, col_names_stitch)
    entropy_1dim_left = np.array(
        [measure_entropy(syn_left.forest.get_tree((ColumnId(i),))) for i in range(len(syn_left.forest.columns))],
        dtype=float,
    )
    entropy_1dim_right = np.array(
        [measure_entropy(syn_right.forest.get_tree((ColumnId(i),))) for i in range(len(syn_right.forest.columns))],
        dtype=float,
    )
    stitching_metadata = StitchingMetadata(
        syn_left.column_is_integral + syn_right.column_is_integral,
        np.concatenate((entropy_1dim_left, entropy_1dim_right)),
    )

    col_names_all = syn_fake.forest.columns
    columns_left = [ColumnId(col_names_all.index(col_name)) for col_name in col_names_left]
    columns_right = [ColumnId(col_names_all.index(col_name)) for col_name in col_names_right]
    columns_right_minus_stitch = [ColumnId(col_names_all.index(col_name)) for col_name in col_names_right_minus_stitch]
    columns_stitch = [ColumnId(col_names_all.index(col_name)) for col_name in col_names_stitch]
    owner = StitchOwner.SHARED if shared else StitchOwner.LEFT
    derived_cluster = (owner, columns_stitch, columns_right_minus_stitch)

    microdata_left = _make_microdata(df_left, syn_left, columns_left)
    microdata_right = _make_microdata(df_right, syn_right, columns_right)

    (microdata, columns) = _do_stitch(
        syn_fake.forest, stitching_metadata, microdata_left, microdata_right, derived_cluster
    )
    col_names = [col_names_all[col_id] for col_id in columns]
    data = [[tup[0] for tup in row] for row in microdata]
    return pd.DataFrame(data, columns=col_names)
